# Standard Library Imports
import logging

from abc import ABCMeta, abstractmethod
from typing import Any

# Protean
from protean.core.entity import BaseEntity
from protean.core.exceptions import (
    ObjectNotFoundError,
    TooManyObjectsError,
    ValidationError,
)
from protean.core.field.basic import Auto, Field
from protean.core.queryset import QuerySet
from protean.globals import current_uow
from protean.utils.query import Q

logger = logging.getLogger("protean.repository")


class ResultSet(object):
    """This is an internal helper class returned by DAO query operations.

    The purpose of this class is to prevent DAO-specific data structures from leaking into the domain layer.
    It can help check whether results exist, traverse the results, fetch the total number of items and also provide
    basic pagination support.
    """

    def __init__(self, offset: int, limit: int, total: int, items: list):
        # the current offset (zero indexed)
        self.offset = offset
        # the number of items to be displayed on a page.
        self.limit = limit
        # the total number of items matching the query
        self.total = total
        # the items for the current page
        self.items = items

    @property
    def has_prev(self):
        """Is `True` if the results are a subset of all results"""
        return bool(self.items) and self.offset > 0

    @property
    def has_next(self):
        """Is `True` if more pages exist"""
        return (self.offset + self.limit) < self.total

    @property
    def first(self):
        """Is the first item from results"""
        if self.items:
            return self.items[0]

    def __bool__(self):
        """Returns `True` when the resultset is not empty"""
        return bool(self.items)

    def __iter__(self):
        """Returns an iterable on items, to support traversal"""
        return iter(self.items)

    def __len__(self):
        """Returns number of items in the resultset"""
        return len(self.items)


class BaseDAO(metaclass=ABCMeta):
    """This is the baseclass for concrete DAO implementations.

    One part of this base class contains abstract methods to be overridden and implemented in each
    concrete database implementation. These methods are where the actual interaction with the database
    takes place. The other part contains fully-implemented object lifecycle methods to help in persistence
    and data fetch. These methods invoke the concrete methods of each database implementation to
    complete their function.

    The lifecycle methods of DAO handle the casting of a model object to a domain entity object, and vice versa.

    :param domain: the domain of the application this DAO is associated with.
    :param provider: the corresponding provider object of the database implementation, from whom the DAO can
                     request and fetch sessions and connections.
    :param model_cls: the concrete model class associated with the DAO. The model class is a direct representation
                      of an object in ORM/ODM terms or as represented by a python driver.
    :param entity_cls: the domain entity class associated with the DAO.
    """

    def __init__(self, domain, provider, entity_cls, model_cls):
        #: Holds a reference to the domain to which the DAO belongs to.
        self.domain = domain

        #: Holds a reference to the provider which supplies the DAO with live connections.
        self.provider = provider

        #: Holds a reference to the model class representation required by the ORM/ODM or the python database driver.
        self.model_cls = model_cls

        #: Holds a reference to the entity class associated with this DAO.
        self.entity_cls = entity_cls

        #: An empty query object that can be used to begin filtering/querying operations
        #: on the underlying data store.
        self.query = QuerySet(self, domain, self.entity_cls)

        #: The actual database document or table name associated with the DAO.
        #: This is used for deriving namespaces for storage containers, especially the default dict provider.
        self.schema_name = entity_cls.meta_.schema_name

        #: Tracks whether the DAO needs to operate outside any active Unit of Work transactions.
        self._outside_uow = False

    def _get_session(self):
        """Returns an active connection to the persistence store.

        - If there is an active transaction, the connection associated with the transaction (in the UoW) is returned
        - If the DAO has been explicitly instructed to work outside a UoW (with the help of `_outside_uow`), or if
          there are no active transactions, a new connection is retrieved from the provider and returned.
        """
        if current_uow and not self._outside_uow:
            return current_uow.get_session(self.provider.name)
        else:
            return self.provider.get_connection()

    def outside_uow(self):
        """When called, the DAO is instructed to work outside active transactions."""
        self._outside_uow = True

        # Return the DAO object to support method chaining
        return self

    ###############################
    # Repository-specific methods #
    ###############################

    @abstractmethod
    def _filter(
        self, criteria: Q, offset: int = 0, limit: int = 10, order_by: list = ()
    ) -> ResultSet:
        """
        Filter objects from the data store. Method must return a `ResultSet`
        object
        """

    @abstractmethod
    def _create(self, model_obj: Any):
        """Persist a new entity into the persistent store. Concrete implementation will be provided by
        the database DAO class.

        Method invocation should persist a new record in the data store.

        This method is invoked by the `create` wrapper and should not be called directly.

        Returns the persisted model object.

        :param model_obj: The model object supplied in an ORM/ODM/Python driver friendly/format
        """

    @abstractmethod
    def _update(self, model_obj: Any):
        """Update entity data in the persistence store. Concrete implementation will be provided by
        the database DAO class.

        Method invocation should update the existing data in the persistent store, by its unique identifier.

        This method is invoked by the DAO's `update` wrapper method and should not be called directly.

        Returns the updated model object.

        :param model_obj: The model object supplied in an ORM/ODM/Python driver friendly/format
        """

    @abstractmethod
    def _update_all(self, criteria: Q, *args, **kwargs):
        """Perform a bulk update on the persistent store. Concrete implementation will be provided by
        the database DAO class.

        Method invocation should update all objects satisfying `criteria` with attributes specified in
        `args` and `kwargs`.

        This method is invoked by Queryset's `update_all()` method and should not be called directly.

        .. warning:: The `update_all()` method is a “bulk” operation, which bypasses ORM/ODM unit-of-work automation
            in favor of greater performance.

        Returns the count of rows matched for the provided criteria.

        :param criteria: A Q object wrapping one or more levels of criteria/filters
        :param args: A dictionary object containing attribute data to be updated
        :param kwargs: Keyword args specifying attribute data to be updated
        """

    @abstractmethod
    def _delete(self):
        """Delete this entity from the persistence store. Concrete implementation will be provided by
        the database DAO class.

        Method invocation should delete existing record in the persistent store, by its unique identifier.

        This method is invoked by DAO's `delete` wrapper method and should not be called directly.

        Returns the deleted model object.
        """

    @abstractmethod
    def _delete_all(self, criteria: Q = None):
        """Perform a bulk delete on the persistent store. Concrete implementation will be provided by
        the database DAO class.

        Method invocation should update all objects satisfying `criteria`.

        This method is invoked by Queryset's `delete_all()` method and should not be called directly.

        .. warning:: The `delete_all()` method is a “bulk” operation, which bypasses ORM/ODM unit-of-work automation
            in favor of greater performance.

        Returns the count of rows matched for the provided criteria.

        :param criteria: A Q object wrapping one or more levels of criteria/filters. If no criteria is provided,
                         then all records of the table/document are removed.
        """

    @abstractmethod
    def _raw(self, query: Any, data: Any = None):
        """Run raw query on Data source. Concrete implementation will be provided by
        the database DAO class.

        Method invocation should fetch all objects satisfying the raw query. Running a raw query on the data store
        should always returns entity instance objects. If the results were not synthesizable back into
        entity objects, an exception should be thrown.

        This method is invoked by Queryset's `raw()` method and should not be called directly.

        .. warning:: The `raw()` method bypasses ORM/ODM unit-of-work automation.

        Returns the result specified by the raw query.

        :param query: Raw query to be passed to the data store.
        :param data: Data to be passed to the data store as-is, along with the query
                     (in case of update statements, for example).
        """

    ######################
    # Life-cycle methods #
    ######################

    def get(self, identifier: Any) -> BaseEntity:
        """Retrieve a specific Record from the Repository by its `identifier`.

        This method internally uses the `filter` method to fetch records.

        Returns exactly one record that matches the identifier.

        Throws `ObjectNotFoundError` if no record was found for the identifier.

        Throws `TooManyObjectsError` if multiple records were found for the identifier.

        :param identifier: id of the record to be fetched from the data store.
        """
        logger.debug(
            f"Lookup `{self.entity_cls.__name__}` object with identifier {identifier}"
        )

        # Filter on the ID field of the entity
        filters = {
            self.entity_cls.meta_.id_field.field_name: identifier,
        }

        results = self.query.filter(**filters).all()
        if not results:
            raise ObjectNotFoundError(
                f"`{self.entity_cls.__name__}` object with identifier {identifier} "
                f"does not exist."
            )

        if len(results) > 1:
            raise TooManyObjectsError(
                f"More than one object of `{self.entity_cls.__name__}` exist with identifier {identifier}",
            )

        # Return the first result, because `filter` would have returned an array
        return results.first

    def find_by(self, **kwargs) -> "BaseEntity":
        """Find a specific entity record that matches one or more criteria.

        This method internally uses the `filter` method to fetch records.

        Returns exactly one record that matches the identifier.

        Throws `ObjectNotFoundError` if no record was found for the identifier.
        Throws `TooManyObjectsError` if multiple records were found for the identifier.

        :param kwargs: named arguments of attribute names and values to filter on.
        """
        logger.debug(
            f"Lookup `{self.entity_cls.__name__}` object with values " f"{kwargs}"
        )

        # Filter for item in the data store
        results = self.query.filter(**kwargs).all()

        if not results:
            raise ObjectNotFoundError(
                f"`{self.entity_cls.__name__}` object with values {[item for item in kwargs.items()]} "
                f"does not exist."
            )

        if len(results) > 1:
            raise TooManyObjectsError(
                f"More than one object of `{self.entity_cls.__name__}` exist "
                f"with values {[item for item in kwargs.items()]}",
            )

        # Return the first result, because `filter` would have returned an array
        return results.first

    def exists(self, excludes_, **filters):
        """Returns `True` if objects matching the provided filters were found. Else, returns False.

        This method internally uses the `filter` method to fetch records. But it can be overridden for better and
        quicker implementations supported by databases.

        :param filters: criteria to match records against
        :param excludes_: criteria that records should not satisfy
        """
        results = self.query.filter(**filters).exclude(**excludes_)

        # Invokes the __bool__ method on `ResultSet`.
        return bool(results)

    def create(self, *args, **kwargs) -> "BaseEntity":
        """Create a new record in the data store.

        Performs validations for unique attributes before creating the entity

        Returns the created entity object.

        Throws `ValidationError` for validation failures on attribute values or uniqueness constraints.

        :param args: Dictionary object containing the object's data.
        :param kwargs: named attribute names and values
        """
        logger.debug(
            f"Creating new `{self.entity_cls.__name__}` object using data {kwargs}"
        )

        try:
            # Build the entity from input arguments
            # Raises validation errors, if any, at this point
            entity_obj = self.entity_cls(*args, **kwargs)

            # Perform unique checks. Raises validation errors if unique constraints are violated.
            self._validate_unique(entity_obj)

            # Build the model object and persist into data store
            model_obj = self._create(self.model_cls.from_entity(entity_obj))

            # Reverse update auto fields into entity
            for field_name, field_obj in entity_obj.meta_.declared_fields.items():
                if isinstance(field_obj, Auto) and not getattr(entity_obj, field_name):
                    if isinstance(model_obj, dict):
                        field_val = model_obj[field_name]
                    else:
                        field_val = getattr(model_obj, field_name)

                    setattr(entity_obj, field_name, field_val)

            # Set Entity status to saved to let everybody know it has been persisted
            entity_obj.state_.mark_saved()

            return entity_obj
        except ValidationError as exc:
            logger.error(f"Failed creating entity because of {exc}")
            raise

    def save(self, entity_obj):
        """Create or update an entity in the data store, depending on its state. An identity for entity record is
        generated, if not already present.

        The primary difference between `save` and other lifecycle methods like `create` and `update` is that `save`
        accepts a fully formed entity object to persist, while the others accept attribute params to build the
        entity model from.

        Returns the created/updated entity object.

        Throws `ValidationError` for validation failures on attribute values or uniqueness constraints.

        :param entity_obj: Entity object to be persisted
        """
        logger.debug(f"Saving `{self.entity_cls.__name__}` object")

        try:
            # Build the model object and create it
            if entity_obj.state_.is_persisted:
                model_obj = self._update(self.model_cls.from_entity(entity_obj))
            else:
                # If this is a new entity, generate ID
                if entity_obj.state_.is_new:
                    if not getattr(
                        entity_obj, entity_obj.meta_.id_field.field_name, None
                    ):
                        setattr(
                            entity_obj,
                            entity_obj.meta_.id_field.field_name,
                            self.entity_cls.generate_identity(),
                        )

                model_obj = self._create(self.model_cls.from_entity(entity_obj))

            updated_entity_obj = self.model_cls.to_entity(model_obj)

            # Update the auto fields of the entity
            for field_name, field_obj in entity_obj.meta_.declared_fields.items():
                if isinstance(field_obj, Auto):
                    field_val = getattr(updated_entity_obj, field_name)
                    setattr(entity_obj, field_name, field_val)

            # Set Entity status to saved to let everybody know it has been persisted
            entity_obj.state_.mark_saved()

            return entity_obj
        except Exception as exc:
            logger.error(f"Failed saving entity because of {exc}")
            raise

    def update(self, entity_obj, *data, **kwargs) -> "BaseEntity":
        """Update a record in the data store.

        Performs validations for unique attributes before creating the entity.

        Supports both dictionary and keyword argument updates to the entity::

            >>> user.update({'age': 10})
            >>> user.update(age=10)

        Returns the updated entity object.

        Throws `ValidationError` for validation failures on attribute values or uniqueness constraints.

        :param entity_obj: The entity object to be updated
        :param data: Dictionary of values to be updated for the entity
        :param kwargs: keyword arguments of attribute pairs to be updated
        """
        logger.debug(
            f"Updating existing `{self.entity_cls.__name__}` object with id {entity_obj.id}"
        )

        try:
            # Update entity's data attributes
            entity_obj._update_data(*data, **kwargs)

            # Do unique checks
            self._validate_unique(entity_obj, create=False)

            self._update(self.model_cls.from_entity(entity_obj))

            # Set Entity status to saved to let everybody know it has been persisted
            entity_obj.state_.mark_saved()

            return entity_obj
        except Exception as exc:
            logger.error(f"Failed updating entity because of {exc}")
            raise

    def _validate_unique(self, entity_obj, create=True):
        """Validate the unique constraints for the entity. Raise ValidationError, if constraints were violated.

        This method internally uses each field object's fail method to construct a valid error message.

        :param entity_obj: Entity object to be validated
        :param create: boolean value to indicate that the validation is part of a create operation
        """
        # Build the filters from the unique constraints
        filters, excludes = {}, {}

        # Construct filter criteria based on unique fields defined in Entity class
        for field_name, field_obj in self.entity_cls.meta_.unique_fields.items():
            lookup_value = getattr(entity_obj, field_name, None)

            # Ignore empty lookup values
            if lookup_value in Field.empty_values:
                continue

            # Ignore identifiers on updates
            if not create and field_obj.identifier:
                excludes[field_name] = lookup_value
                continue

            filters[field_name] = lookup_value

        # Lookup the objects by filters and raise error if objects exist
        for filter_key, lookup_value in filters.items():
            if self.exists(excludes, **{filter_key: lookup_value}):
                field_obj = self.entity_cls.meta_.declared_fields[filter_key]
                field_obj.fail(
                    "unique",
                    entity_name=self.entity_cls.__name__,
                    field_name=filter_key,
                )

    def delete(self, entity_obj):
        """Delete a record in the data store.

        Performs validations before data deletion.

        Returns the deleted entity object.

        Throws ObjectNotFoundError if the object was not found in the data store.

        :param entity_obj: Entity object to be deleted from data store
        """
        try:
            if not entity_obj.state_.is_destroyed:
                self._delete(self.model_cls.from_entity(entity_obj))

                # Set Entity status to destroyed to let everybody know the object is no longer referable
                entity_obj.state_.mark_destroyed()

            return entity_obj
        except Exception as exc:
            logger.error(f"Failed entity deletion because of {exc}")
            raise

    def delete_all(self):
        """Delete all records in this table/document in the persistent store.

        Does not perform validations before data deletion.
        Does not return confirmation of data deletion.
        """
        try:
            self._delete_all()
        except Exception as exc:
            logger.error(f"Failed deletion of all records because of {exc}")
            raise


class BaseLookup(metaclass=ABCMeta):
    """Base Lookup class to implement for each lookup

    Inspired by the lookup mechanism implemented in Django.

    Each lookup, which is simply a data comparison (like `name == 'John'`), is implemented as a subclass of this
    class, and has to implement the `as_expression()` method to provide the representation that the persistence
    store needs.

    Lookups are identified by their names, and the names are stored in the `lookup_name` class variable.
    """

    lookup_name = None

    def __init__(self, source, target):
        """Source is LHS and Target is RHS of a comparsion.

        For example, in the expression `name == 'John'`, `name` is source (LHS) and `'John'` is target (RHS).
        In other words, source is the key/column/attribute to be searched on, and target is the value present in the
        persistent store.
        """
        self.source, self.target = source, target

    def process_source(self):
        """This is a blank implementation that simply returns the source.

        Returns `source` (LHS of the expression).

        You can override this method to manipulate the source when necessary. For example, if you are using a
        data store that cannot perform case-insensitive queries, it may be useful to always compare in lowercase.
        """
        return self.source

    def process_target(self):
        """This is a blank implementation that simply returns the target.

        Returns `target` (RHS of the expression).

        You can override this method to manipulate the target when necessary. A good example of overriding this
        method is when you are using a data store that needs strings to be enclosed in single quotes.
        """
        return self.target

    @abstractmethod
    def as_expression(self):
        """This methods should return the source and the target in the format required by the persistence store.

        Concrete implementation for this method varies from database to database.
        """
        raise NotImplementedError



# Protean
from protean.port.dao import ResultSet


class TestResultSet:
    def test_successful_resultset_initialization(self):
        resultset = ResultSet(offset=0, limit=10, total=2, items=["foo", "bar"])

        assert resultset is not None

    def test_resultset_properties(self):
        resultset = ResultSet(offset=0, limit=10, total=2, items=["foo", "bar"])

        assert len(resultset) == 2
        assert resultset.first == "foo"
        assert [item for item in resultset] == ["foo", "bar"]  # Test __iter__

    def test_pagination_properties_of_resultset(self):
        resultset = ResultSet(offset=0, limit=10, total=2, items=["foo", "bar"])
        has_next_resultset = ResultSet(offset=0, limit=2, total=4, items=["foo", "bar"])
        has_prev_resultset = ResultSet(offset=2, limit=2, total=4, items=["foo", "bar"])

        assert resultset.has_prev is False
        assert resultset.has_next is False
        assert has_next_resultset.has_next is True
        assert has_next_resultset.has_prev is False
        assert has_prev_resultset.has_next is False
        assert has_prev_resultset.has_prev is True

    def test_boolean_evaluation_of_resultset(self):
        resultset = ResultSet(offset=0, limit=10, total=2, items=["foo", "bar"])
        empty_resultset = ResultSet(offset=0, limit=10, total=0, items=[])

        assert bool(resultset) is True
        assert bool(empty_resultset) is False



# Protean
import pytest

from mock import patch
from protean.core.command_handler import BaseCommandHandler
from protean.utils import fully_qualified_name

# Local/Relative Imports
from .elements import AddNewPersonCommandHandler, AddPersonCommand


class TestCommandHandlerInitialization:
    def test_that_base_command_handler_class_cannot_be_instantiated(self):
        with pytest.raises(TypeError):
            BaseCommandHandler()

    def test_that_command_handler_can_be_instantiated(self, test_domain):
        service = AddNewPersonCommandHandler(
            test_domain, AddPersonCommand(first_name="John", last_name="Doe", age=21)
        )
        assert service is not None


class TestCommandHandlerRegistration:
    def test_that_command_handler_can_be_registered_with_domain(self, test_domain):
        test_domain.register(AddNewPersonCommandHandler)

        assert (
            fully_qualified_name(AddNewPersonCommandHandler)
            in test_domain.registry.command_handlers
        )

    def test_that_command_handler_can_be_registered_via_annotations(self, test_domain):
        @test_domain.command_handler(command=AddPersonCommand)
        class AnnotatedCommandHandler:
            def special_method(self):
                pass

        assert (
            fully_qualified_name(AnnotatedCommandHandler)
            in test_domain.registry.command_handlers
        )


class TestDomainEventNotification:
    @patch.object(AddNewPersonCommandHandler, "notify")
    def test_that_domain_event_is_received_from_aggregate_command_method(
        self, mock, test_domain
    ):
        test_domain.register(AddNewPersonCommandHandler)

        command = AddPersonCommand(first_name="John", last_name="Doe", age=21)
        test_domain.publish_command(command)
        mock.assert_called_once_with(command.to_dict())



# Standard Library Imports
from collections import defaultdict
from enum import Enum

# Protean
from protean.core.aggregate import BaseAggregate
from protean.core.field.basic import Float, Identifier, Integer, String
from protean.core.field.embedded import ValueObjectField
from protean.core.value_object import BaseValueObject


class Email(BaseValueObject):
    """An email address value object, with two clearly identified parts:
        * local_part
        * domain_part
    """

    # This is the external facing data attribute
    address = String(max_length=254, required=True)

    @classmethod
    def from_address(cls, address):
        """ Construct an Email VO from an email address.

        email = Email.from_address('john.doe@gmail.com')

        """
        if not cls.validate(address):
            raise ValueError("Email address is invalid")

        return cls(address=address)

    @classmethod
    def validate(cls, address):
        """ Business rules of Email address """
        if (
            type(address) is not str
            or "@" not in address
            or address.count("@") > 1
            or len(address) > 255
        ):
            return False

        return True


class User(BaseAggregate):
    email = ValueObjectField(Email, required=True)
    name = String(max_length=255)


class MyOrgEmail(Email):
    pass


class Currency(Enum):
    """ Set of choices for the status"""

    USD = "USD"
    INR = "INR"
    CAD = "CAD"


class Balance(BaseValueObject):
    """A composite amount object, containing two parts:
        * currency code - a three letter unique currency code
        * amount - a float value
    """

    currency = String(max_length=3, choices=Currency)
    amount = Float()

    def clean(self):
        errors = defaultdict(list)
        if self.amount and self.amount < -1000000000000.0:
            errors["amount"].append("cannot be less than 1 Trillion")
        return errors

    def replace(self, **kwargs):
        # FIXME Find a way to do this generically and move method to `BaseValueObject`
        currency = kwargs.pop("currency", None)
        amount = kwargs.pop("amount", None)
        return Balance(currency=currency or self.currency, amount=amount or self.amount)


class Account(BaseAggregate):
    balance = ValueObjectField(Balance, required=True)
    kind = String(max_length=15, required=True)


class BuildingStatus(Enum):
    WIP = "WIP"
    DONE = "DONE"


class Building(BaseValueObject):
    name = String(max_length=50)
    floors = Integer()
    status = String(choices=BuildingStatus)

    def defaults(self):
        if not self.status:
            if self.floors == 4:
                self.status = BuildingStatus.DONE.value
            else:
                self.status = BuildingStatus.WIP.value

    def clean(self):
        errors = defaultdict(list)
        if self.floors >= 4 and self.status != BuildingStatus.DONE.value:
            errors["status"].append("should be DONE")
        return errors


class PolymorphicConnection(BaseValueObject):
    connected_id = Identifier(referenced_as="connected_id")
    connected_type = String(referenced_as="connected_type", max_length=15)


class PolymorphicOwner(BaseAggregate):
    connector = ValueObjectField(PolymorphicConnection)



# Protean
import pytest

from protean.core.exceptions import NotSupportedError

# Local/Relative Imports
from .elements import AbstractRole, ConcreteRole


class TestAggregateAbstraction:
    def test_that_abstract_entities_cannot_be_registered_or_initialized(
        self, test_domain
    ):
        with pytest.raises(NotSupportedError) as exc1:
            test_domain.register(AbstractRole)
        assert exc1.value.args[0] == (
            "AbstractRole class has been marked abstract" " and cannot be instantiated"
        )

        with pytest.raises(NotSupportedError) as exc2:
            AbstractRole(name="Titan")
        assert exc2.value.args[0] == (
            "AbstractRole class has been marked abstract" " and cannot be instantiated"
        )

    def test_that_concrete_entities_can_be_created_from_abstract_entities_through_inheritance(
        self, test_domain
    ):
        test_domain.register(ConcreteRole)
        concrete_role = ConcreteRole(name="Titan")
        assert concrete_role is not None
        assert concrete_role.name == "Titan"

    def test_that_abstract_entities_can_be_created_with_annotations(self, test_domain):
        # Protean
        from protean.core.aggregate import BaseAggregate
        from protean.core.field.basic import String

        class CustomBaseClass(BaseAggregate):
            foo = String(max_length=25)

            class Meta:
                abstract = True

        @test_domain.aggregate
        class ConcreateSubclass(CustomBaseClass):
            bar = String(max_length=25)

        assert all(
            key in ConcreateSubclass.meta_.declared_fields for key in ["foo", "bar"]
        )

        concrete = ConcreateSubclass(foo="Saturn", bar="Titan")
        assert concrete is not None
        assert concrete.foo == "Saturn"
        assert concrete.bar == "Titan"



# Protean
import pytest

from protean.core.exceptions import ValidationError

# Local/Relative Imports
from .elements import Building, BuildingStatus


class TestDefaults:
    def test_that_building_is_marked_as_done_if_above_4_floors(self):
        building = Building(name="Foo", floors=4)

        assert building.status == BuildingStatus.DONE.value

    def test_that_building_is_marked_as_done_if_below_4_floors(self):
        building = Building(name="Foo", floors=1)

        assert building.status == BuildingStatus.WIP.value


class TestClean:
    def test_that_building_cannot_be_WIP_if_above_4_floors(self):
        with pytest.raises(ValidationError):
            Building(name="Foo", floors=4, status=BuildingStatus.WIP.value)



"""Entity Functionality and Classes"""
# Standard Library Imports
import copy
import logging

from collections import defaultdict
from uuid import uuid4

# Protean
from protean.core.exceptions import (
    ConfigurationError,
    IncorrectUsageError,
    NotSupportedError,
    ValidationError,
)
from protean.core.field.association import Association, Reference
from protean.core.field.basic import Auto, Field
from protean.core.field.embedded import ValueObjectField
from protean.globals import current_domain
from protean.utils import (
    DomainObjects,
    IdentityStrategy,
    IdentityType,
    derive_element_class,
    inflection,
)

# Local/Relative Imports
from ..core.field.association import _ReferenceField  # Relative path to private class

logger = logging.getLogger("protean.domain.entity")


class _EntityMetaclass(type):
    """
    This base metaclass processes the class declaration and constructs a meta object that can
    be used to introspect the Entity class later. Specifically, it sets up a `meta_` attribute on
    the Entity to an instance of Meta, either the default of one that is defined in the
    Entity class.

    `meta_` is setup with these attributes:
        * `declared_fields`: A dictionary that gives a list of any instances of `Field`
            included as attributes on either the class or on any of its superclasses
        * `id_field`: The Primary identifier attribute of the Entity
    """

    def __new__(mcs, name, bases, attrs, **kwargs):
        """Initialize Entity MetaClass and load attributes"""

        # Ensure initialization is only performed for subclasses of Entity
        # (excluding Entity class itself).
        parents = [b for b in bases if isinstance(b, _EntityMetaclass)]
        if not parents:
            return super().__new__(mcs, name, bases, attrs)

        # Remove `abstract` if defined in base classes
        for base in bases:
            if hasattr(base, "Meta") and hasattr(base.Meta, "abstract"):
                delattr(base.Meta, "abstract")

        new_class = super().__new__(mcs, name, bases, attrs, **kwargs)

        # Gather `Meta` class/object if defined
        attr_meta = attrs.pop(
            "Meta", None
        )  # Gather Metadata defined in inner `Meta` class
        entity_meta = EntityMeta(name, attr_meta)  # Initialize the Metadata container
        setattr(
            new_class, "meta_", entity_meta
        )  # Associate the Metadata container with new class

        # Load declared fields
        new_class._load_fields(attrs)

        # Load declared fields from Base class, in case this Entity is subclassing another
        new_class._load_base_class_fields(bases, attrs)

        # Lookup an already defined ID field or create an `Auto` field
        new_class._set_id_field()

        # Set up Relation Fields
        new_class._set_up_reference_fields()

        # Set up ValueObject Fields
        new_class._set_up_value_object_fields()

        return new_class

    def _load_base_class_fields(new_class, bases, attrs):
        """If this class is subclassing another Entity, add that Entity's
        fields.  Note that we loop over the bases in *reverse*.
        This is necessary in order to maintain the correct order of fields.
        """
        for base in reversed(bases):
            if hasattr(base, "meta_") and hasattr(base.meta_, "declared_fields"):
                base_class_fields = {
                    field_name: field_obj
                    for (field_name, field_obj) in base.meta_.declared_fields.items()
                    if (
                        field_name not in attrs
                        and not isinstance(field_obj, Association)
                        and not field_obj.identifier
                    )
                }
                new_class._load_fields(base_class_fields)

    def _load_fields(new_class, attrs):
        """Load field items into Class.

        This method sets up the primary attribute of an association.
        If Child class has defined an attribute so `parent = field.Reference(Parent)`, then `parent`
        is set up in this method, while `parent_id` is set up in `_set_up_reference_fields()`.
        """
        for attr_name, attr_obj in attrs.items():
            if isinstance(attr_obj, (Association, Field, Reference)):
                setattr(new_class, attr_name, attr_obj)
                new_class.meta_.declared_fields[attr_name] = attr_obj

    def _set_up_reference_fields(new_class):
        """Walk through relation fields and setup shadow attributes"""
        if new_class.meta_.declared_fields:
            for _, field in new_class.meta_.declared_fields.items():
                if isinstance(field, Reference):
                    shadow_field_name, shadow_field = field.get_shadow_field()
                    new_class.meta_.reference_fields[shadow_field_name] = shadow_field
                    shadow_field.__set_name__(new_class, shadow_field_name)

    def _set_up_value_object_fields(new_class):
        """Walk through value object fields and setup shadow attributes"""
        if new_class.meta_.declared_fields:
            for _, field in new_class.meta_.declared_fields.items():
                if isinstance(field, ValueObjectField):
                    shadow_fields = field.get_shadow_fields()
                    for shadow_field_name, shadow_field in shadow_fields:
                        new_class.meta_.value_object_fields[
                            shadow_field_name
                        ] = shadow_field

    def _set_id_field(new_class):
        """Lookup the id field for this entity and assign"""
        # FIXME What does it mean when there are no declared fields?
        #   Does it translate to an abstract entity?
        if new_class.meta_.declared_fields:
            try:
                new_class.meta_.id_field = next(
                    field
                    for _, field in new_class.meta_.declared_fields.items()
                    if isinstance(field, (Field, Reference)) and field.identifier
                )
            except StopIteration:
                # If no id field is declared then create one
                new_class._create_id_field()

    def _create_id_field(new_class):
        """Create and return a default ID field that is Auto generated"""
        id_field = Auto(identifier=True)

        setattr(new_class, "id", id_field)
        id_field.__set_name__(new_class, "id")

        # Ensure ID field is updated properly in Meta attribute
        new_class.meta_.declared_fields["id"] = id_field
        new_class.meta_.id_field = id_field


class EntityMeta:
    """ Metadata info for the entity.

    Options:
    - ``abstract``: Indicates that this is an abstract entity (Ignores all other meta options)
    - ``schema_name``: name of the schema (table/index/doc) used for persistence of this entity
        defaults to underscore version of the Entity name.
    - ``provider``: the name of the datasource associated with this
        entity, default value is `default`.
    - ``order_by``: default ordering of objects returned by filter queries.

    Also acts as a placeholder for generated entity fields like:

        :declared_fields: dict
            Any instances of `Field` included as attributes on either the class
            or on any of its superclasses will be include in this dictionary.
        :id_field: protean.core.Field
            An instance of the field that will serve as the unique identifier for the entity

    FIXME Make `EntityMeta` immutable
    """

    def __init__(self, entity_name, meta):
        self.abstract = getattr(meta, "abstract", None) or False
        self.schema_name = getattr(meta, "schema_name", None) or inflection.underscore(
            entity_name
        )
        self.provider = getattr(meta, "provider", None) or "default"
        self.model = getattr(meta, "model", None)

        # `order_by` can be provided either as a string or a tuple
        ordering = getattr(meta, "order_by", ())
        if isinstance(ordering, str):
            self.order_by = (ordering,)
        else:
            self.order_by = tuple(ordering)

        # Initialize Options
        self.declared_fields = {}
        self.value_object_fields = {}
        self.reference_fields = {}
        self.id_field = None

        # Domain Attributes
        self.aggregate_cls = getattr(meta, "aggregate_cls", None)

    @property
    def mandatory_fields(self):
        """ Return the mandatory fields for this entity """
        return {
            field_name: field_obj
            for field_name, field_obj in self.attributes.items()
            if not isinstance(field_obj, Association) and field_obj.required
        }

    @property
    def unique_fields(self):
        """ Return the unique fields for this entity """
        return {
            field_name: field_obj
            for field_name, field_obj in self.attributes.items()
            if not isinstance(field_obj, Association) and field_obj.unique
        }

    @property
    def auto_fields(self):
        return {
            field_name: field_obj
            for field_name, field_obj in self.declared_fields.items()
            if isinstance(field_obj, Auto)
        }

    @property
    def attributes(self):
        attributes_dict = {}
        for _, field_obj in self.declared_fields.items():
            if isinstance(field_obj, ValueObjectField):
                shadow_fields = field_obj.get_shadow_fields()
                for _, shadow_field in shadow_fields:
                    attributes_dict[shadow_field.attribute_name] = shadow_field
            elif isinstance(field_obj, Reference):
                attributes_dict[field_obj.get_attribute_name()] = field_obj.relation
            elif isinstance(field_obj, Field):
                attributes_dict[field_obj.get_attribute_name()] = field_obj
            else:  # This field is an association. Ignore recording it as an attribute
                pass

        return attributes_dict


class _FieldsCacheDescriptor:
    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        res = instance.fields_cache = {}
        return res


class _EntityState:
    """Store entity instance state."""

    def __init__(self):
        self._new = True
        self._changed = False
        self._destroyed = False

    @property
    def is_new(self):
        return self._new

    @property
    def is_persisted(self):
        return not self._new

    @property
    def is_changed(self):
        return self._changed

    @property
    def is_destroyed(self):
        return self._destroyed

    def mark_new(self):
        self._new = True

    def mark_saved(self):
        self._new = False
        self._changed = False

    mark_retrieved = (
        mark_saved  # Alias as placeholder so that future change wont affect interface
    )

    def mark_changed(self):
        if not (self._new or self._destroyed):
            self._changed = True

    def mark_destroyed(self):
        self._destroyed = True
        self._changed = False

    fields_cache = _FieldsCacheDescriptor()


class BaseEntity(metaclass=_EntityMetaclass):
    """The Base class for Protean-Compliant Domain Entities.

    Provides helper methods to custom define entity attributes, and query attribute names
    during runtime.

    You can define an Entity with the help of `entity` annotation::

        @domain.entity
        class User:
            id = field.Integer(identifier=True)
            first_name = field.String(required=True, max_length=50)
            last_name = field.String(required=True, max_length=50)
            age = field.Integer(default=5)

    (or)

    Or, you can directly subclass from `BaseEntity`::

        class User(BaseEntity):
            id = field.Integer(identifier=True)
            first_name = field.String(required=True, max_length=50)
            last_name = field.String(required=True, max_length=50)
            age = field.Integer(default=5)

        domain.register_element(User)

    During persistence, the model associated with this entity is retrieved dynamically from the repository factory,
    initialized with a live connection to the datastore.
    """

    element_type = DomainObjects.ENTITY

    def __init__(self, *template, raise_errors=True, **kwargs):  # noqa: C901
        """
        Initialise the entity object.

        During initialization, set value on fields if validation passes.

        This initialization technique supports keyword arguments as well as dictionaries. The objects initialized
        in the following example have the same structure::

            user1 = User({'first_name': 'John', 'last_name': 'Doe'})

            user2 = User(first_name='John', last_name='Doe')

        You can also specify a template for initial data and override specific attributes::

            base_user = User({'age': 15})

            user = User(base_user.to_dict(), first_name='John', last_name='Doe')
        """

        if self.meta_.abstract is True:
            raise NotSupportedError(
                f"{self.__class__.__name__} class has been marked abstract"
                f" and cannot be instantiated"
            )

        self.errors = defaultdict(list)
        self.raise_errors = raise_errors

        # Set up the storage for instance state
        self.state_ = _EntityState()

        # Load the attributes based on the template
        loaded_fields = []
        for dictionary in template:
            if not isinstance(dictionary, dict):
                raise AssertionError(
                    f'Positional argument "{dictionary}" passed must be a dict.'
                    f"This argument serves as a template for loading common "
                    f"values.",
                )
            for field_name, val in dictionary.items():
                if field_name not in kwargs:
                    kwargs[field_name] = val

        # Now load against the keyword arguments
        for field_name, val in kwargs.items():
            loaded_fields.append(field_name)
            try:
                setattr(self, field_name, val)
            except ValidationError as err:
                for field_name in err.messages:
                    self.errors[field_name].extend(err.messages[field_name])

        # Load Value Objects from associated fields
        #   This block will dynamically construct value objects from field values
        #   and associated the vo with the entity
        # If the value object was already provided, it will not be overridden.
        for field_name, field_obj in self.meta_.declared_fields.items():
            if isinstance(field_obj, (ValueObjectField)) and not getattr(
                self, field_name
            ):
                attributes = [
                    (embedded_field.field_name, embedded_field.attribute_name)
                    for embedded_field in field_obj.embedded_fields.values()
                ]
                values = {name: kwargs.get(attr) for name, attr in attributes}
                try:
                    value_object = field_obj.value_object_cls.build(**values)
                    # Set VO value only if the value object is not None/Empty
                    if value_object:
                        setattr(self, field_name, value_object)
                        loaded_fields.append(field_name)
                except ValidationError as err:
                    for sub_field_name in err.messages:
                        self.errors["{}_{}".format(field_name, sub_field_name)].extend(
                            err.messages[sub_field_name]
                        )

        # Load Identities
        if (
            not getattr(self, self.meta_.id_field.field_name, None)
            and type(self.meta_.id_field) is Auto
        ):
            setattr(self, self.meta_.id_field.field_name, self.generate_identity())
            loaded_fields.append(self.meta_.id_field.field_name)

        # Load Associations
        for field_name, field_obj in self.meta_.declared_fields.items():
            if isinstance(field_obj, Association):
                getattr(self, field_name)  # This refreshes the values in associations

        # Now load the remaining fields with a None value, which will fail
        # for required fields
        for field_name, field_obj in self.meta_.declared_fields.items():
            if field_name not in loaded_fields:
                if not isinstance(field_obj, (Reference, _ReferenceField, Association)):
                    try:
                        setattr(self, field_name, None)

                        # If field is a VO, set underlying attributes to None as well
                        if isinstance(field_obj, ValueObjectField):
                            for embedded_field in field_obj.embedded_fields.values():
                                setattr(self, embedded_field.attribute_name, None)
                    except ValidationError as err:
                        for field_name in err.messages:
                            self.errors[field_name].extend(err.messages[field_name])

        for field_name, field_obj in self.meta_.attributes.items():
            if field_name not in loaded_fields and not hasattr(self, field_name):
                setattr(self, field_name, None)

        self.defaults()

        # `clean()` will return a `defaultdict(list)` if errors are to be raised
        custom_errors = self.clean() or {}
        for field in custom_errors:
            self.errors[field].extend(custom_errors[field])

        # Raise any errors found during load
        if self.errors and self.raise_errors:
            logger.error(self.errors)
            raise ValidationError(self.errors)

    def defaults(self):
        """Placeholder method for defaults.
        To be overridden in concrete Containers, when an attribute's default depends on other attribute values.
        """

    def clean(self):
        """Placeholder method for validations.
        To be overridden in concrete Containers, when complex validations spanning multiple fields are required.
        """
        return defaultdict(list)

    @classmethod
    def generate_identity(cls):
        """Generate Unique Identifier, based on configured strategy"""
        if current_domain.config["IDENTITY_STRATEGY"] == IdentityStrategy.UUID:
            if current_domain.config["IDENTITY_TYPE"] == IdentityType.INTEGER:
                return uuid4().int
            elif current_domain.config["IDENTITY_TYPE"] == IdentityType.STRING:
                return str(uuid4())
            elif current_domain.config["IDENTITY_TYPE"] == IdentityType.UUID:
                return uuid4()
            else:
                raise ConfigurationError(
                    f'Unknown Identity Type {current_domain.config["IDENTITY_TYPE"]}'
                )

        return None  # Database will generate the identity

    def __eq__(self, other):
        """Equivalence check to be based only on Identity"""

        # FIXME Enhanced Equality Checks
        #   * Ensure IDs have values and both of them are not null
        #   * Ensure that the ID is of the right type
        #   * Ensure that Objects belong to the same `type`
        #   * Check Reference equality

        # FIXME Check if `==` and `in` operator work with __eq__

        if type(other) is type(self):
            self_id = getattr(self, self.meta_.id_field.field_name)
            other_id = getattr(other, other.meta_.id_field.field_name)

            return self_id == other_id

        return False

    def __hash__(self):
        """Overrides the default implementation and bases hashing on identity"""

        # FIXME Add Object Class Type to hash
        return hash(getattr(self, self.meta_.id_field.field_name))

    def _update_data(self, *data_dict, **kwargs):
        """
        A private method to process and update entity values correctly.

        :param data: A dictionary of values to be updated for the entity
        :param kwargs: keyword arguments with key-value pairs to be updated
        """

        # Load each of the fields given in the data dictionary
        self.errors = {}

        for data in data_dict:
            if not isinstance(data, dict):
                raise AssertionError(
                    f'Positional argument "{data}" passed must be a dict.'
                    f"This argument serves as a template for loading common "
                    f"values.",
                )
            for field_name, val in data.items():
                setattr(self, field_name, val)

        # Now load against the keyword arguments
        for field_name, val in kwargs.items():
            setattr(self, field_name, val)

        # Raise any errors found during update
        if self.errors:
            logger.error(self.errors)
            raise ValidationError(self.errors)

    def to_dict(self):
        """ Return entity data as a dictionary """
        field_values = {}
        field_values = {
            field_name: getattr(self, field_name, None)
            for field_name, field_obj in self.meta_.declared_fields.items()
            if not isinstance(field_obj, ValueObjectField)
        }

        # FIXME Simplify fetching and appending Value Object dict values
        vo_fields = {
            field_name: getattr(self, field_name, None)
            for field_name, field_obj in self.meta_.declared_fields.items()
            if isinstance(field_obj, ValueObjectField)
        }

        vo_field_values = {}
        for vo_field_name, vo_field_obj in vo_fields.items():
            if vo_field_obj:
                vo_field_values.update(
                    {
                        vo_field_name
                        + "_"
                        + field_name: getattr(vo_field_obj, field_name, None)
                        for field_name, field_obj in vo_field_obj.meta_.declared_fields.items()
                    }
                )

        return {**field_values, **vo_field_values}

    def __repr__(self):
        """Friendly repr for Entity"""
        return "<%s: %s>" % (self.__class__.__name__, self)

    def __str__(self):
        identifier = getattr(self, self.meta_.id_field.field_name)
        return "%s object (%s)" % (
            self.__class__.__name__,
            "{}: {}".format(self.meta_.id_field.field_name, identifier),
        )

    def clone(self):
        """Deepclone the entity, but reset state"""
        clone_copy = copy.deepcopy(self)
        clone_copy.state_ = _EntityState()

        return clone_copy


def entity_factory(element_cls, **kwargs):
    element_cls = derive_element_class(element_cls, BaseEntity)

    if element_cls.meta_.abstract is True:
        raise NotSupportedError(
            f"{element_cls.__name__} class has been marked abstract"
            f" and cannot be instantiated"
        )

    element_cls.meta_.provider = (
        kwargs.pop("provider", None)
        or (hasattr(element_cls, "meta_") and element_cls.meta_.provider)
        or "default"
    )
    element_cls.meta_.model = (
        kwargs.pop("model", None)
        or (hasattr(element_cls, "meta_") and element_cls.meta_.model)
        or None
    )
    element_cls.meta_.aggregate_cls = (
        kwargs.pop("aggregate_cls", None)
        or (hasattr(element_cls, "meta_") and element_cls.meta_.aggregate_cls)
        or None
    )

    if not element_cls.meta_.aggregate_cls:
        raise IncorrectUsageError(
            f"Entity `{element_cls.__name__}` needs to be associated with an Aggregate"
        )

    return element_cls



# Protean
from protean.core.aggregate import BaseAggregate
from protean.core.field.basic import Integer, String
from protean.core.serializer import BaseSerializer


class User(BaseAggregate):
    name = String(required=True)
    age = Integer(required=True)


class UserSchema(BaseSerializer):
    name = String(required=True)
    age = Integer(required=True)

    class Meta:
        aggregate_cls = User



# Standard Library Imports
import logging

# Protean
from protean.utils import DomainObjects, derive_element_class

logger = logging.getLogger("protean.application")


class BaseApplicationService:
    """Base ApplicationService class that all other Application services should inherit from.

    This class is a placeholder class for now. Application concepts directly influence the
    method names in concrete Application Service classes, so no abstract methods are necessary.
    Each Application Service class is usually associated one-to-one with API calls.

    Application services are responsible for fetching the linked domain, initializing repositories,
    caches, and message brokers, and injecting dependencies into the domain layer. These are automatable
    aspects that can be part of the base class in the future.
    """

    element_type = DomainObjects.APPLICATION_SERVICE

    def __new__(cls, *args, **kwargs):
        if cls is BaseApplicationService:
            raise TypeError("BaseApplicationService cannot be instantiated")
        return object.__new__(cls, *args, **kwargs)


def application_service_factory(element_cls, **kwargs):
    return derive_element_class(element_cls, BaseApplicationService)



""" Module defines utilities for importing modules and packages """

# Standard Library Imports
from importlib import import_module


def perform_import(val):
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    """
    if val is not None:
        if isinstance(val, str):
            return import_from_string(val)
        elif isinstance(val, (list, tuple)):
            return [import_from_string(item) for item in val]

    return val


def import_from_string(val):
    """
    Attempt to import a class from a string representation.
    """
    try:
        module_path, class_name = val.rsplit(".", 1)
        module = import_module(module_path)
        return getattr(module, class_name)
    except (ImportError, AttributeError) as e:
        msg = f"Could not import {val}. {e.__class__.__name__}: {e}"
        raise ImportError(msg)



# Protean
import mock
import pytest

from protean.core.queryset import QuerySet
from protean.port.dao import ResultSet

# Local/Relative Imports
from .elements import (
    Account,
    AccountVia,
    AccountViaWithReference,
    Author,
    Comment,
    CommentVia,
    CommentViaWithReference,
    Post,
    PostVia,
    PostViaWithReference,
    Profile,
    ProfileVia,
    ProfileViaWithReference,
)


class TestHasOne:
    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Account)
        test_domain.register(Author)
        test_domain.register(AccountVia)
        test_domain.register(AccountViaWithReference)
        test_domain.register(Post)
        test_domain.register(Profile)
        test_domain.register(ProfileVia)
        test_domain.register(ProfileViaWithReference)

    def test_successful_initialization_of_entity_with_has_one_association(
        self, test_domain
    ):
        account = Account(email="john.doe@gmail.com", password="a1b2c3")
        test_domain.get_dao(Account).save(account)
        author = Author(first_name="John", last_name="Doe", account=account)
        test_domain.get_dao(Author).save(author)

        assert all(key in author.__dict__ for key in ["account", "account_email"])
        assert author.account.email == account.email
        assert author.account_email == account.email

        refreshed_account = test_domain.get_dao(Account).get(account.email)
        assert refreshed_account.author.id == author.id
        assert refreshed_account.author == author

    def test_successful_has_one_initialization_with_a_class_containing_via_and_no_reference(
        self, test_domain
    ):
        account = AccountVia(email="john.doe@gmail.com", password="a1b2c3")
        test_domain.get_dao(AccountVia).save(account)
        profile = ProfileVia(
            profile_id="12345", about_me="Lorem Ipsum", account_email=account.email
        )
        test_domain.get_dao(ProfileVia).save(profile)

        refreshed_account = test_domain.get_dao(AccountVia).get(account.email)
        assert refreshed_account.profile == profile

    def test_successful_has_one_initialization_with_a_class_containing_via_and_reference(
        self, test_domain
    ):
        account = AccountViaWithReference(
            email="john.doe@gmail.com", password="a1b2c3", username="johndoe"
        )
        test_domain.get_dao(AccountViaWithReference).save(account)
        profile = ProfileViaWithReference(about_me="Lorem Ipsum", ac=account)
        test_domain.get_dao(ProfileViaWithReference).save(profile)

        refreshed_account = test_domain.get_dao(AccountViaWithReference).get(
            account.email
        )
        assert refreshed_account.profile == profile

    @mock.patch("protean.core.field.association.Association._fetch_objects")
    def test_that_subsequent_access_after_first_retrieval_do_not_fetch_record_again(
        self, mock, test_domain
    ):
        account = AccountViaWithReference(
            email="john.doe@gmail.com", password="a1b2c3", username="johndoe"
        )
        test_domain.get_dao(AccountViaWithReference).save(account)
        profile = ProfileViaWithReference(about_me="Lorem Ipsum", ac=account)
        test_domain.get_dao(ProfileViaWithReference).save(profile)

        mock.return_value = profile

        refreshed_account = test_domain.get_dao(AccountViaWithReference).get(
            account.email
        )
        for _ in range(3):
            getattr(refreshed_account, "profile")
        assert (
            mock.call_count == 0
        )  # This is because `profile` would have been loaded when account was fetched


class TestHasMany:
    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Post)
        test_domain.register(PostVia)
        test_domain.register(PostViaWithReference)
        test_domain.register(Comment)
        test_domain.register(CommentVia)
        test_domain.register(CommentViaWithReference)

    @pytest.fixture
    def persisted_post(self, test_domain):
        post = test_domain.get_dao(Post).create(
            title="Test Post", slug="test-post", content="Do Re Mi Fa"
        )
        return post

    def test_successful_initialization_of_entity_with_has_many_association(
        self, test_domain
    ):
        post = Post(content="Lorem Ipsum")
        test_domain.repository_for(Post).add(post)

        comment1 = Comment(id=101, content="First Comment")
        comment2 = Comment(id=102, content="Second Comment")

        post.comments.add(comment1)
        post.comments.add(comment2)
        test_domain.repository_for(Post).add(post)

        refreshed_post = test_domain.repository_for(Post).get(post.id)
        assert len(refreshed_post.comments) == 2
        assert "comments" in refreshed_post.__dict__  # Available after access
        assert refreshed_post.comments[0].post_id == post.id
        assert refreshed_post.comments[1].post_id == post.id

        assert isinstance(refreshed_post.comments, QuerySet)
        assert isinstance(refreshed_post.comments.all(), ResultSet)
        assert all(
            comment.id in [101, 102] for comment in refreshed_post.comments
        )  # `__iter__` magic here

    def test_successful_has_one_initialization_with_a_class_containing_via_and_no_reference(
        self, test_domain
    ):
        post = PostVia(content="Lorem Ipsum")
        test_domain.get_dao(PostVia).save(post)
        comment1 = CommentVia(id=101, content="First Comment", posting_id=post.id)
        comment2 = CommentVia(id=102, content="First Comment", posting_id=post.id)
        test_domain.get_dao(CommentVia).save(comment1)
        test_domain.get_dao(CommentVia).save(comment2)

        assert comment1.posting_id == post.id
        assert comment2.posting_id == post.id

        refreshed_post = test_domain.get_dao(PostVia).get(post.id)
        assert len(refreshed_post.comments) == 2
        assert "comments" in refreshed_post.__dict__  # Available after access

        assert isinstance(refreshed_post.comments, QuerySet)
        assert isinstance(refreshed_post.comments.all(), ResultSet)
        assert all(
            comment.id in [101, 102] for comment in refreshed_post.comments
        )  # `__iter__` magic here

    def test_successful_has_one_initialization_with_a_class_containing_via_and_reference(
        self, test_domain
    ):
        post = PostViaWithReference(content="Lorem Ipsum")
        test_domain.get_dao(PostViaWithReference).save(post)
        comment1 = CommentViaWithReference(
            id=101, content="First Comment", posting=post
        )
        comment2 = CommentViaWithReference(
            id=102, content="First Comment", posting=post
        )
        test_domain.get_dao(CommentViaWithReference).save(comment1)
        test_domain.get_dao(CommentViaWithReference).save(comment2)

        assert comment1.posting_id == post.id
        assert comment2.posting_id == post.id

        refreshed_post = test_domain.get_dao(PostViaWithReference).get(post.id)
        assert len(refreshed_post.comments) == 2
        assert "comments" in refreshed_post.__dict__  # Available after access

        assert isinstance(refreshed_post.comments, QuerySet)
        assert isinstance(refreshed_post.comments.all(), ResultSet)
        assert all(
            comment.id in [101, 102] for comment in refreshed_post.comments
        )  # `__iter__` magic here

    def test_that_subsequent_access_after_first_retrieval_do_not_fetch_record_again(
        self, test_domain
    ):
        post = PostViaWithReference(content="Lorem Ipsum")
        test_domain.get_dao(PostViaWithReference).save(post)
        comment1 = CommentViaWithReference(
            id=101, content="First Comment", posting=post
        )
        comment2 = CommentViaWithReference(
            id=102, content="First Comment", posting=post
        )
        test_domain.get_dao(CommentViaWithReference).save(comment1)
        test_domain.get_dao(CommentViaWithReference).save(comment2)

        refreshed_post = test_domain.get_dao(PostViaWithReference).get(post.id)
        with mock.patch(
            "protean.core.field.association.HasMany._fetch_objects"
        ) as mock_fetch_objects:
            for _ in range(3):
                getattr(refreshed_post, "comments")
        assert mock_fetch_objects.call_count == 0

    def test_that_entities_up_to_configured_limit_value_are_retrieved(
        self, test_domain, persisted_post
    ):
        for i in range(1, 13):
            comment = Comment(
                content=f"Comment {i}", post_id=persisted_post.id
            )  # FIXME This should not be necessary
            test_domain.get_dao(Comment).save(comment)
            persisted_post.comments.add(comment)
            test_domain.get_dao(Post).save(persisted_post)

        updated_post = test_domain.get_dao(Post).get(persisted_post.id)
        assert updated_post.comments.total == 12
        assert len(updated_post.comments.items) == 12

    def test_that_entities_beyond_configured_limit_value_are_not_retrieved(
        self, test_domain, persisted_post
    ):
        for i in range(1, 20):
            comment = Comment(
                content=f"Comment {i}", post_id=persisted_post.id
            )  # FIXME This should not be necessary
            test_domain.get_dao(Comment).save(comment)

        updated_post = test_domain.get_dao(Post).get(persisted_post.id)
        assert updated_post.comments.total == 19
        assert len(updated_post.comments.items) == 15

    def test_filtering_on_has_many_association(self, test_domain, persisted_post):
        comments = []

        for i in range(1, 13):
            comment = Comment(
                content=f"Comment {i}", post_id=persisted_post.id
            )  # FIXME This should not be necessary
            comments.append(comment)
            test_domain.get_dao(Comment).save(comment)

        post = test_domain.get_dao(Post).get(persisted_post.id)
        specific_comment = post.comments.filter(content="Comment 2").all()
        assert specific_comment.total == 1
        assert specific_comment.items[0] == comments[1]



# Standard Library Imports
import logging

# Protean
from protean.utils import DomainObjects, derive_element_class

logger = logging.getLogger("protean.domain.service")


class BaseDomainService:
    """Base DomainService class that all other domain services should inherit from.

    This is a placeholder class for now. Methods that are implemented
    in concreate Domain Service classes are inspired from Domain concepts,
    and typically use more than one aggregate to accomplish a task"""

    element_type = DomainObjects.DOMAIN_SERVICE

    def __new__(cls, *args, **kwargs):
        if cls is BaseDomainService:
            raise TypeError("BaseDomainService cannot be instantiated")
        return object.__new__(cls, *args, **kwargs)


def domain_service_factory(element_cls, **kwargs):
    return derive_element_class(element_cls, BaseDomainService)



# Standard Library Imports
import os

# Protean
import pytest

from elasticsearch_dsl import Index


def initialize_domain():
    # Protean
    from protean.domain import Domain

    domain = Domain("Elasticsearch Tests")

    # Construct relative path to config file
    current_path = os.path.abspath(os.path.dirname(__file__))
    config_path = os.path.join(current_path, "./config.py")

    if os.path.exists(config_path):
        domain.config.from_pyfile(config_path)

    return domain


@pytest.fixture
def test_domain():
    domain = initialize_domain()
    with domain.domain_context():
        yield domain


@pytest.fixture(scope="session", autouse=True)
def setup_db():
    domain = initialize_domain()
    with domain.domain_context():
        # Create all indexes
        # Local/Relative Imports
        from .elements import Alien, ComplexUser, Person, Provider, User

        domain.register(Person)
        domain.register(Alien)
        domain.register(User)
        domain.register(ComplexUser)
        domain.register(Provider)

        provider = domain.get_provider("default")
        conn = provider.get_connection()

        for _, aggregate_record in domain.registry.aggregates.items():
            index = Index(aggregate_record.cls.meta_.schema_name, using=conn)
            if not index.exists():
                index.create()

        yield

        # Drop all indexes at the end of test suite
        for _, aggregate_record in domain.registry.aggregates.items():
            index = Index(aggregate_record.cls.meta_.schema_name, using=conn)
            if index.exists():
                index.delete()


@pytest.fixture(autouse=True)
def run_around_tests(test_domain):
    yield
    if test_domain.providers.has_provider("default"):
        test_domain.get_provider("default")._data_reset()



# Protean
from click.testing import CliRunner
from protean.cli import main


def test_main():
    runner = CliRunner()
    result = runner.invoke(main, ["--version"])

    # Protean
    from protean import __version__

    assert result.output == "main, version {protean_version}\n".format(
        protean_version=__version__
    )
    assert result.exit_code == 0



# Standard Library Imports
from abc import abstractmethod

# Protean
from protean.core.exceptions import IncorrectUsageError
from protean.utils import DomainObjects, derive_element_class


class ModelMeta:
    """ Metadata info for the Model.

    Options:
    - ``entity_cls``: The Entity that this model is associated with
    """

    def __init__(self, meta=None):
        if meta:
            self.entity_cls = getattr(meta, "entity_cls", None)
            self.schema = getattr(meta, "schema", None)
            self.database = getattr(meta, "database", None)
        else:
            self.entity_cls = None
            self.schema = None
            self.database = None


class BaseModel:
    """This is a Model representing a data schema in the persistence store. A concrete implementation of this
    model has to be provided by each persistence store plugin.
    """

    element_type = DomainObjects.MODEL

    def __new__(cls, *args, **kwargs):
        if cls is BaseModel:
            raise TypeError("BaseModel cannot be instantiated")
        return super().__new__(cls)

    @classmethod
    @abstractmethod
    def from_entity(cls, entity):
        """Initialize Model object from Entity object"""

    @classmethod
    @abstractmethod
    def to_entity(cls, *args, **kwargs):
        """Convert Model Object to Entity Object"""


def model_factory(element_cls, **kwargs):
    element_cls = derive_element_class(element_cls, BaseModel)

    if hasattr(element_cls, "Meta"):
        element_cls.meta_ = ModelMeta(element_cls.Meta)
    else:
        element_cls.meta_ = ModelMeta()

    if not (hasattr(element_cls.meta_, "entity_cls") and element_cls.meta_.entity_cls):
        element_cls.meta_.entity_cls = kwargs.pop("entity_cls", None)

    if not (hasattr(element_cls.meta_, "schema") and element_cls.meta_.schema):
        element_cls.meta_.schema = kwargs.pop("schema", None)

    if not (hasattr(element_cls.meta_, "database") and element_cls.meta_.database):
        element_cls.meta_.database = kwargs.pop("database", None)

    if not element_cls.meta_.entity_cls:
        raise IncorrectUsageError(
            "Models need to be associated with an Entity or Aggregate"
        )

    return element_cls



# Standard Library Imports
from datetime import datetime
from uuid import UUID, uuid4

# Protean
import pytest

from protean.core.exceptions import InvalidOperationError, ValidationError
from protean.core.field.basic import Auto, String

# Local/Relative Imports
from .elements import PersonAutoSSN, PersonExplicitID, Role, SubclassRole


class TestProperties:
    def test_conversion_of_aggregate_values_to_dict(self):
        current_time = datetime.now()
        role = Role(id=12, name="ADMIN", created_on=current_time)
        assert role.to_dict() == {"id": 12, "name": "ADMIN", "created_on": current_time}

    def test_repr_output_of_aggregate(self):
        role = Role(id=12, name="ADMIN")

        assert str(role) == "Role object (id: 12)"
        assert repr(role) == "<Role: Role object (id: 12)>"


class TestIdentity:
    """Grouping of Identity related test cases"""

    def test_id_field_in_meta(self):
        assert hasattr(Role.meta_, "id_field")
        assert type(Role.meta_.id_field) is Auto

    def test_default_id_field_construction(self):
        assert "id" in Role.meta_.declared_fields
        assert "id" in Role.meta_.attributes

        assert type(Role.meta_.declared_fields["id"]) is Auto
        assert Role.meta_.id_field == Role.meta_.declared_fields["id"]

    def test_non_default_auto_id_field_construction(self):
        assert "id" not in PersonAutoSSN.meta_.declared_fields
        assert "id" not in PersonAutoSSN.meta_.attributes

        assert type(PersonAutoSSN.meta_.declared_fields["ssn"]) is Auto
        assert PersonAutoSSN.meta_.id_field.field_name == "ssn"
        assert (
            PersonAutoSSN.meta_.id_field == PersonAutoSSN.meta_.declared_fields["ssn"]
        )

    def test_non_default_explicit_id_field_construction(self, test_domain):
        assert "id" not in PersonExplicitID.meta_.declared_fields
        assert "id" not in PersonExplicitID.meta_.attributes

        assert type(PersonExplicitID.meta_.declared_fields["ssn"]) is String
        assert PersonExplicitID.meta_.id_field.field_name == "ssn"
        assert (
            PersonExplicitID.meta_.id_field
            == PersonExplicitID.meta_.declared_fields["ssn"]
        )


class TestIdentityValues:
    """Grouping of Identity value related test cases"""

    def test_default_id_field_value(self):
        role = Role(name="ADMIN")
        assert role.id is not None

        try:
            uuid_obj = UUID(str(role.id))
        except ValueError:
            pytest.fail("ID is not valid UUID")

        assert str(uuid_obj) == role.id

    def test_that_ids_are_immutable(self):
        """Test that `id` cannot be changed once assigned"""
        role = Role(id=12, name="ADMIN")

        with pytest.raises(InvalidOperationError):
            role.id = 13

    def test_non_default_auto_id_field_value(self):
        role = PersonAutoSSN(name="John Doe")
        assert role.ssn is not None

        try:
            uuid_obj = UUID(str(role.ssn))
        except ValueError:
            pytest.fail("SSN is not valid UUID")

        assert str(uuid_obj) == role.ssn

    def test_non_default_explicit_id_field_value(self):
        with pytest.raises(ValidationError):
            role = PersonExplicitID(name="John Doe")

        new_uuid = uuid4()
        role = PersonExplicitID(ssn=new_uuid, name="John Doe")
        assert role is not None
        assert role.ssn == str(new_uuid)

    def test_that_explicit_id_can_be_supplied_to_auto_id_field(self):
        new_uuid = uuid4()
        role = Role(id=new_uuid, name="ADMIN")
        assert role.id is not None
        assert role.id == new_uuid


class TestEquivalence:
    def test_that_two_entities_with_same_id_are_treated_as_equal(self):
        role1 = Role(id=12345, name="ADMIN1")
        role2 = Role(id=12346, name="ADMIN1")

        assert role1 != role2  # Because their identities are different
        assert role2 != role1  # Because their identities are different

        role3 = Role(id=12345, name="ADMIN1")
        assert (
            role1 == role3
        )  # Because it's the same record even though attributes differ
        assert role3 == role1

    def test_that_two_entities_of_different_types_are_different_even_with_same_id(self):
        """Test that two entities are not considered equal even if they have the same ID
            and one belongs to a different Entity class
        """
        role = Role(id=1, name="ADMIN")
        person = PersonAutoSSN(id=1, name="ADMIN")

        assert role != person  # Even though their identities are the same
        assert person != role  # Even though their identities are the same

    def test_that_two_entities_of_inherited_types_are_different_even_with_same_id(self):
        """Test that two entities are not considered equal even if they have the same ID
            and one is subclassed from the other
        """
        role = Role(id=1, name="ADMIN")
        subclass_role = SubclassRole(id=1, name="ADMIN")

        assert role != subclass_role  # Even though their identities are the same
        assert subclass_role != role  # Even though their identities are the same

    def test_generated_aggregate_hash(self):
        """Test that the entity's hash is based on its identity"""
        hashed_id = hash(1)

        role = Role(id=1, name="ADMIN")
        assert hashed_id == hash(
            role
        )  # FIXME Should hash be based on ID alone, or other attrs too?

    def test_that_two_aggregates_that_are_equal_have_equal_hash(self):
        role1 = Role(id=12345, name="ADMIN1")
        role2 = Role(id=12345, name="ADMIN1")

        assert hash(role1) == hash(role2)



# Standard Library Imports
from datetime import datetime
from typing import List

# Protean
from protean.core.aggregate import BaseAggregate
from protean.core.field.basic import DateTime, Integer, String
from protean.core.repository import BaseRepository


class Person(BaseAggregate):
    first_name = String(max_length=50, required=True)
    last_name = String(max_length=50, required=True)
    age = Integer(default=21)
    created_at = DateTime(default=datetime.now())


class PersonRepository(BaseRepository):
    def find_adults(self, age: int = 21) -> List[Person]:
        pass  # FIXME Implement filter method


class User(BaseAggregate):
    email = String(max_length=255, required=True, unique=True)
    password = String(max_length=3026)



# Standard Library Imports
from datetime import datetime

# Protean
import pytest

from protean.core.aggregate import BaseAggregate
from protean.core.field.basic import DateTime, Dict, String


class Event(BaseAggregate):
    name = String(max_length=255)
    created_at = DateTime(default=datetime.utcnow())
    payload = Dict()


@pytest.mark.postgresql
def test_persistence_and_retrieval(test_domain):
    test_domain.register(Event)

    repo = test_domain.repository_for(Event)
    event = Event(
        name="UserCreated", payload={"email": "john.doe@gmail.com", "password": "*****"}
    )
    repo.add(event)

    event_dup = test_domain.get_dao(Event).find_by(name="UserCreated")
    assert event_dup is not None
    assert event_dup.payload is not None
    assert event_dup.payload == {"email": "john.doe@gmail.com", "password": "*****"}



"""Utility classes and methods for DB Adapters, Repositories and Query Constructors"""
# Standard Library Imports
import copy
import functools
import inspect
import logging

logger = logging.getLogger("protean.repository")


def subclasses(cls):
    """Iterator utilty to loop and clear registered Lookups against a class"""
    yield cls
    for subclass in cls.__subclasses__():
        yield from subclasses(subclass)


class RegisterLookupMixin:
    """Helper Mixin to register Lookups to an Adapter"""

    @classmethod
    def _get_lookup(cls, lookup_name):
        return cls.get_lookups().get(lookup_name, None)

    @classmethod
    @functools.lru_cache(maxsize=None)
    def get_lookups(cls):
        """Fetch all Lookups"""
        class_lookups = [
            parent.__dict__.get("class_lookups", {}) for parent in inspect.getmro(cls)
        ]
        return cls.merge_dicts(class_lookups)

    def get_lookup(self, lookup_name):
        """Fetch Lookup by name"""
        # Protean
        from protean.port.dao import BaseLookup

        lookup = self._get_lookup(lookup_name)

        # If unable to find Lookup, or if Lookup is the wrong class, raise Error
        if lookup is None or (
            lookup is not None and not issubclass(lookup, BaseLookup)
        ):
            raise NotImplementedError

        return lookup

    @staticmethod
    def merge_dicts(dicts):
        """
        Merge dicts in reverse to preference the order of the original list. e.g.,
        merge_dicts([a, b]) will preference the keys in 'a' over those in 'b'.
        """
        merged = {}
        for d in reversed(dicts):
            merged.update(d)
        return merged

    @classmethod
    def _clear_cached_lookups(cls):
        for subclass in subclasses(cls):
            subclass.get_lookups.cache_clear()

    @classmethod
    def register_lookup(cls, lookup, lookup_name=None):
        """Register a Lookup to a class"""
        if lookup_name is None:
            lookup_name = lookup.lookup_name
        if "class_lookups" not in cls.__dict__:
            cls.class_lookups = {}

        cls.class_lookups[lookup_name] = lookup
        cls._clear_cached_lookups()

        return lookup

    @classmethod
    def _delist_lookup(cls, lookup, lookup_name=None):
        """
        Remove given lookup from cls lookups. For use in tests only as it's
        not thread-safe.
        """
        if lookup_name is None:
            lookup_name = lookup.lookup_name
        del cls.class_lookups[lookup_name]


class Node:
    """
    A class for storing a tree graph. Primarily used for filter constructs.

    A single internal node in the tree graph. A Node should be viewed as a
    connection (the root) with the children being either leaf nodes or other
    Node instances.
    """

    # Standard connector type. Clients usually won't use this at all and
    # subclasses will usually override the value.
    default = "DEFAULT"

    def __init__(self, children=None, connector=None, negated=False):
        """Construct a new Node. If no connector is given, use the default."""
        self.children = children[:] if children else []
        self.connector = connector or self.default
        self.negated = negated

    # Required because django.db.models.query_utils.Q. Q. __init__() is
    # problematic, but it is a natural Node subclass in all other respects.
    @classmethod
    def _new_instance(cls, children=None, connector=None, negated=False):
        """
        Create a new instance of this class when new Nodes (or subclasses) are
        needed in the internal code in this class. Normally, it just shadows
        __init__(). However, subclasses with an __init__ signature that aren't
        an extension of Node.__init__ might need to implement this method to
        allow a Node to create a new instance of them (if they have any extra
        setting up to do).
        """
        obj = Node(children, connector, negated)
        obj.__class__ = cls
        return obj

    def __str__(self):
        template = "(NOT (%s: %s))" if self.negated else "(%s: %s)"
        return template % (self.connector, ", ".join(str(c) for c in self.children))

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, self)

    def __deepcopy__(self, memodict):
        obj = Node(connector=self.connector, negated=self.negated)
        obj.__class__ = self.__class__
        obj.children = copy.deepcopy(self.children, memodict)
        return obj

    def __len__(self):
        """Return the number of children this node has."""
        return len(self.children)

    def __bool__(self):
        """Return whether or not this node has children."""
        return bool(self.children)

    def __contains__(self, other):
        """Return True if 'other' is a direct child of this instance."""
        return other in self.children

    def __eq__(self, other):
        return (
            self.__class__ == other.__class__
            and (self.connector, self.negated) == (other.connector, other.negated)
            and self.children == other.children
        )

    def add(self, data, conn_type, squash=True):
        """
        Combine this tree and the data represented by data using the
        connector conn_type. The combine is done by squashing the node other
        away if possible.

        This tree (self) will never be pushed to a child node of the
        combined tree, nor will the connector or negated properties change.

        Return a node which can be used in place of data regardless if the
        node other got squashed or not.

        If `squash` is False the data is prepared and added as a child to
        this tree without further logic.
        """
        if data in self.children:
            return data
        if not squash:
            self.children.append(data)
            return data
        if self.connector == conn_type:
            # We can reuse self.children to append or squash the node other.
            if (
                isinstance(data, Node)
                and not data.negated
                and (data.connector == conn_type or len(data) == 1)
            ):
                # We can squash the other node's children directly into this
                # node. We are just doing (AB)(CD) == (ABCD) here, with the
                # addition that if the length of the other node is 1 the
                # connector doesn't matter. However, for the len(self) == 1
                # case we don't want to do the squashing, as it would alter
                # self.connector.
                self.children.extend(data.children)
                return self
            else:
                # We could use perhaps additional logic here to see if some
                # children could be used for pushdown here.
                self.children.append(data)
                return data
        else:
            obj = self._new_instance(self.children, self.connector, self.negated)
            self.connector = conn_type
            self.children = [obj, data]
            return data

    def negate(self):
        """Negate the sense of the root connector."""
        self.negated = not self.negated


class Q(Node):
    """
    Encapsulate filters as objects that can then be combined logically (using
    `&` and `|`).
    """

    # Connection types
    AND = "AND"
    OR = "OR"
    default = AND
    conditional = True

    def __init__(self, *args, _connector=None, _negated=False, **kwargs):
        super().__init__(
            children=[*args, *sorted(kwargs.items())],
            connector=_connector,
            negated=_negated,
        )

    def _combine(self, other, conn):
        if not isinstance(other, Q):
            raise TypeError(other)

        # If the other Q() is empty, ignore it and just use `self`.
        if not other:
            return copy.deepcopy(self)
        # Or if this Q is empty, ignore it and just use `other`.
        elif not self:
            return copy.deepcopy(other)

        obj = type(self)()
        obj.connector = conn
        obj.add(self, conn)
        obj.add(other, conn)
        return obj

    def __or__(self, other):
        return self._combine(other, self.OR)

    def __and__(self, other):
        return self._combine(other, self.AND)

    def __invert__(self):
        obj = type(self)()
        obj.add(self, self.AND)
        obj.negate()
        return obj

    def deconstruct(self):
        """Deconstruct a Q Object"""
        path = "%s.%s" % (self.__class__.__module__, self.__class__.__name__)
        args, kwargs = (), {}

        if len(self.children) == 1 and not isinstance(self.children[0], Q):
            child = self.children[0]
            kwargs = {child[0]: child[1]}
        else:
            args = tuple(self.children)
            if self.connector != self.default:
                kwargs = {"_connector": self.connector}
        if self.negated:
            kwargs["_negated"] = True
        return path, args, kwargs



# Protean
import pytest

from protean.adapters.repository.memory import MemoryModel

# Local/Relative Imports
from .elements import Email, Person, Provider, ProviderCustomModel, Receiver, User


class TestModel:
    @pytest.fixture(autouse=True)
    def register_person_aggregate(self, test_domain):
        test_domain.register(Person)

    def test_that_model_class_is_created_automatically(self, test_domain):
        model_cls = test_domain.get_model(Person)
        assert issubclass(model_cls, MemoryModel)
        assert model_cls.__name__ == "PersonModel"

    def test_conversation_from_entity_to_model(self, test_domain):
        model_cls = test_domain.get_model(Person)

        person = Person(first_name="John", last_name="Doe")

        person_model_obj = model_cls.from_entity(person)

        assert person_model_obj is not None
        assert isinstance(person_model_obj, dict)
        assert person_model_obj["age"] == 21
        assert person_model_obj["first_name"] == "John"
        assert person_model_obj["last_name"] == "Doe"

    def test_conversation_from_model_to_entity(self, test_domain):
        model_cls = test_domain.get_model(Person)
        person = Person(first_name="John", last_name="Doe")
        person_model_obj = model_cls.from_entity(person)

        person_copy = model_cls.to_entity(person_model_obj)
        assert person_copy is not None


class TestModelWithVO:
    @pytest.fixture(autouse=True)
    def register_user_aggregate(self, test_domain):
        test_domain.register(User)

    def test_that_model_class_is_created_automatically(self, test_domain):
        model_cls = test_domain.get_model(User)
        assert issubclass(model_cls, MemoryModel)
        assert model_cls.__name__ == "UserModel"

    def test_conversation_from_entity_to_model(self, test_domain):
        model_cls = test_domain.get_model(User)

        user1 = User(email_address="john.doe@gmail.com", password="d4e5r6")
        user2 = User(email=Email(address="john.doe@gmail.com"), password="d4e5r6")

        user1_model_obj = model_cls.from_entity(user1)
        user2_model_obj = model_cls.from_entity(user2)

        assert user1_model_obj is not None
        assert isinstance(user1_model_obj, dict)
        assert user1_model_obj["email_address"] == "john.doe@gmail.com"
        assert user1_model_obj["password"] == "d4e5r6"

        assert user2_model_obj is not None
        assert isinstance(user2_model_obj, dict)
        assert user2_model_obj["email_address"] == "john.doe@gmail.com"
        assert user2_model_obj["password"] == "d4e5r6"

        # Model's content should reflect only the attributes, not declared_fields
        assert "email" not in user1_model_obj
        assert "email" not in user2_model_obj

    def test_conversion_from_model_to_entity(self, test_domain):
        model_cls = test_domain.get_model(User)
        user1 = User(email_address="john.doe@gmail.com", password="d4e5r6")
        user1_model_obj = model_cls.from_entity(user1)

        user_copy = model_cls.to_entity(user1_model_obj)
        assert user_copy is not None
        assert user_copy.id == user1_model_obj["id"]


class TestCustomModel:
    def test_that_custom_model_is_associated_with_entity(self, test_domain):
        test_domain.register(Provider)
        test_domain.register(ProviderCustomModel)

        model_cls = test_domain.get_model(Provider)
        assert model_cls.__name__ == "ProviderCustomModel"

    def test_that_model_can_be_registered_with_domain_annotation(self, test_domain):
        # Protean
        from protean.core.field.basic import Text

        test_domain.register(Receiver)

        @test_domain.model(entity_cls=Receiver)
        class ReceiverInlineModel:
            about = Text()

        model_cls = test_domain.get_model(Receiver)
        assert model_cls.__name__ == "ReceiverInlineModel"

        # FIXME This test will fail in the future
        #   when models are validated for fields to be present in corresponding entities/aggregates
        assert hasattr(model_cls, "about")



# Standard Library Imports
import os

# Protean
import pytest


def initialize_domain():
    # Protean
    from protean.domain import Domain

    domain = Domain("SQLAlchemy Test - SQLite")

    # Construct relative path to config file
    current_path = os.path.abspath(os.path.dirname(__file__))
    config_path = os.path.join(current_path, "./config.py")

    if os.path.exists(config_path):
        domain.config.from_pyfile(config_path)

    return domain


domain = initialize_domain()


@pytest.fixture(autouse=True)
def test_domain():
    with domain.domain_context():
        yield domain


@pytest.fixture(scope="session", autouse=True)
def setup_db():
    with domain.domain_context():
        # Create all associated tables
        # Local/Relative Imports
        from .elements import ComplexUser, Person, Provider, ProviderCustomModel, User
        from .test_array_datatype import ArrayUser, IntegerArrayUser
        from .test_json_datatype import Event

        domain.register(ArrayUser)
        domain.register(ComplexUser)
        domain.register(Event)
        domain.register(IntegerArrayUser)
        domain.register(Person)
        domain.register(Provider)
        domain.register(User)
        domain.register_model(ProviderCustomModel, entity_cls=Provider)

        domain.get_dao(ArrayUser)
        domain.get_dao(ComplexUser)
        domain.get_dao(Event)
        domain.get_dao(IntegerArrayUser)
        domain.get_dao(Person)
        domain.get_dao(Provider)
        domain.get_dao(User)

        for provider in domain.providers_list():
            provider._metadata.create_all()

        yield

        # Drop all tables at the end of test suite
        for provider in domain.providers_list():
            provider._metadata.drop_all()


@pytest.fixture(autouse=True)
def run_around_tests(test_domain):
    yield
    if test_domain.providers.has_provider("default"):
        test_domain.get_provider("default")._data_reset()



# Standard Library Imports
from abc import abstractmethod

# Protean
from protean.utils import DomainObjects, derive_element_class


class _CommandHandlerMetaclass(type):
    """
    This base metaclass processes the class declaration and constructs a meta object that can
    be used to introspect the CommandHandler class later. Specifically, it sets up a `meta_` attribute on
    the CommandHandler to an instance of Meta, either the default of one that is defined in the
    CommandHandler class.

    `meta_` is setup with these attributes:
        * `command`: The command that this command handler is associated with
    """

    def __new__(mcs, name, bases, attrs, **kwargs):
        """Initialize CommandHandler MetaClass and load attributes"""

        # Ensure initialization is only performed for subclasses of CommandHandler
        # (excluding CommandHandler class itself).
        parents = [b for b in bases if isinstance(b, _CommandHandlerMetaclass)]
        if not parents:
            return super().__new__(mcs, name, bases, attrs)

        # Remove `abstract` in base classes if defined
        for base in bases:
            if hasattr(base, "Meta") and hasattr(base.Meta, "abstract"):
                delattr(base.Meta, "abstract")

        new_class = super().__new__(mcs, name, bases, attrs, **kwargs)

        # Gather `Meta` class/object if defined
        attr_meta = attrs.pop("Meta", None)
        meta = attr_meta or getattr(new_class, "Meta", None)
        setattr(new_class, "meta_", CommandHandlerMeta(name, meta))

        return new_class


class CommandHandlerMeta:
    """ Metadata info for the CommandHandler.

    Options:
    - ``command``: The command that this command handler is associated with
    """

    def __init__(self, entity_name, meta):
        self.command_cls = getattr(meta, "command_cls", None)
        self.broker = getattr(meta, "broker", "default")


class BaseCommandHandler(metaclass=_CommandHandlerMetaclass):
    """Base Subsciber class that should implemented by all Domain CommandHandlers.

    This is also a marker class that is referenced when command handlers are registered
    with the domain
    """

    element_type = DomainObjects.COMMAND_HANDLER

    def __new__(cls, *args, **kwargs):
        if cls is BaseCommandHandler:
            raise TypeError("BaseCommandHandler cannot be instantiated")
        return super().__new__(cls)

    def __init__(self, domain, command_cls):
        self.domain = domain
        self.command_cls = command_cls

    @abstractmethod
    def notify(self, command):
        """Placeholder method for receiving notifications on command"""
        pass


def command_handler_factory(element_cls, **kwargs):
    return derive_element_class(element_cls, BaseCommandHandler)



# Protean
from protean.utils import Database, IdentityStrategy, IdentityType

DEBUG = True
TESTING = True
ENV = "development"
AGGREGATE_CHILDREN_LIMIT = 15

# A secret key for this particular Protean installation. Used in secret-key
# hashing algorithms.
SECRET_KEY = "tvTpk3PAfkGr5x9!2sFU%XpW7bR8cwKA"

# Database Configuration
DATABASES = {
    "default": {"PROVIDER": "protean.adapters.MemoryProvider"},
    "sqlite": {
        "PROVIDER": "protean.adapters.repository.sqlalchemy.SAProvider",
        "DATABASE": Database.SQLITE.value,
        "DATABASE_URI": "sqlite:///test.db",
    },
}

# Identity strategy to use when persisting Entities/Aggregates.
#
# Options:
#
#   * IdentityStrategy.UUID: Default option, and preferred. Identity is a UUID and generated during `build` time.
#       Persisted along with other details into the data store.
#   * IdentityStrategy.DATABASE: Let the database generate unique identity during persistence
#   * IdentityStrategy.FUNCTION: Special function that returns a unique identifier
IDENTITY_STRATEGY = IdentityStrategy.UUID

# Data type of Auto-Generated Identity Values
#
# Options:
#
#   * INTEGER
#   * STRING (Default)
IDENTITY_TYPE = IdentityType.STRING

# Messaging Mediums
BROKERS = {
    "default": {"PROVIDER": "protean.adapters.InlineBroker"},
}

EMAIL_PROVIDERS = {
    "default": {
        "PROVIDER": "protean.adapters.DummyEmailProvider",
        "DEFAULT_FROM_EMAIL": "admin@team8solutions.com",
    },
}



# Protean
import pytest

from protean.core.application_service import BaseApplicationService
from protean.utils import fully_qualified_name

# Local/Relative Imports
from .elements import DummyApplicationService


class TestApplicationServiceInitialization:
    def test_that_base_application_service_class_cannot_be_instantiated(self):
        with pytest.raises(TypeError):
            BaseApplicationService()

    def test_that_application_service_can_be_instantiated(self):
        service = DummyApplicationService()
        assert service is not None


class TestApplicationServiceRegistration:
    def test_that_application_service_can_be_registered_with_domain(self, test_domain):
        test_domain.register(DummyApplicationService)

        assert (
            fully_qualified_name(DummyApplicationService)
            in test_domain.registry.application_services
        )

    def test_that_application_service_can_be_registered_via_annotations(
        self, test_domain
    ):
        @test_domain.application_service
        class AnnotatedApplicationService:
            def special_method(self):
                pass

        assert (
            fully_qualified_name(AnnotatedApplicationService)
            in test_domain.registry.application_services
        )



# Standard Library Imports
import logging
import logging.config

from abc import abstractmethod
from collections import defaultdict
from collections.abc import Iterable

# Protean
from protean.utils import DomainObjects, fully_qualified_name

logger = logging.getLogger("protean.port.broker")


class _BrokerMetaclass(type):
    """
    This base metaclass processes the class declaration and constructs a meta object that can
    be used to introspect the Broker class later. Specifically, it sets up a `meta_` attribute on
    the Broker to an instance of Meta, either the default of one that is defined in the
    Broker class.

    `meta_` is setup with these attributes:
        * `aggregate`: The aggregate associated with the repository
    """

    def __new__(mcs, name, bases, attrs, **kwargs):
        """Initialize Broker MetaClass and load attributes"""

        # Ensure initialization is only performed for subclasses of Broker
        # (excluding Broker class itself).
        parents = [b for b in bases if isinstance(b, _BrokerMetaclass)]
        if not parents:
            return super().__new__(mcs, name, bases, attrs)

        # Remove `abstract` in base classes if defined
        for base in bases:
            if hasattr(base, "Meta") and hasattr(base.Meta, "abstract"):
                delattr(base.Meta, "abstract")

        new_class = super().__new__(mcs, name, bases, attrs, **kwargs)

        # Gather `Meta` class/object if defined
        attr_meta = attrs.pop("Meta", None)
        meta = attr_meta or getattr(new_class, "Meta", None)
        setattr(new_class, "meta_", BrokerMeta(name, meta))

        return new_class


class BrokerMeta:
    """ Metadata info for the Broker.

    Options:
    - ``aggregate_cls``: The aggregate associated with the repository
    """

    def __init__(self, entity_name, meta):
        self.aggregate_cls = getattr(meta, "aggregate_cls", None)


class BaseBroker(metaclass=_BrokerMetaclass):
    """This class outlines the base broker functions,
    to be satisfied by all implementing brokers.

    It is also a marker interface for registering broker
    classes with the domain"""

    def __init__(self, name, domain, conn_info):
        self.name = name
        self.domain = domain
        self.conn_info = conn_info

        self._subscribers = defaultdict(set)
        self._command_handlers = {}

    @abstractmethod
    def get_connection(self):
        """Get the connection object to the broker"""

    @abstractmethod
    def send_message(self, initiator_obj):
        """Placeholder method for brokers to accept incoming events"""

    def register(self, initiator_cls, consumer_cls):
        """Registers Domain Events and Commands with Subscribers/Command Handlers

        Arguments:
            initiator_cls {list} -- One or more Domain Events or Commands
            consumer_cls {Subscriber/CommandHandler} -- The consumer class connected to the Domain Event or Command
        """
        if not isinstance(initiator_cls, Iterable):
            initiator_cls = [initiator_cls]

        for initiator in initiator_cls:
            if initiator.element_type == DomainObjects.DOMAIN_EVENT:
                self._subscribers[fully_qualified_name(initiator)].add(consumer_cls)
                logger.debug(
                    f"Registered Subscriber {consumer_cls.__name__} with broker {self.name}"
                )
            else:
                self._command_handlers[fully_qualified_name(initiator)] = consumer_cls



# Protean
import pytest

from mock import patch
from protean.adapters import SendgridEmailProvider

# Local/Relative Imports
from .elements import Person, PersonAdded, WelcomeEmail, WelcomeNewPerson


@pytest.mark.sendgrid
class TestEmailTriggering:
    @patch.object(SendgridEmailProvider, "send_email")
    def test_that_email_is_pushed_via_aggregate_command_method(self, mock, test_domain):
        test_domain.register(PersonAdded)
        test_domain.register(WelcomeEmail)
        test_domain.register(WelcomeNewPerson)

        Person.add_newcomer(
            {
                "email": "john.doe@gmail.com",
                "first_name": "John",
                "last_name": "Doe",
                "age": 21,
            }
        )
        mock.assert_called_once()

    def test_that_sendgrid_email_method_is_called(self, mocker, test_domain):
        test_domain.register(PersonAdded)
        test_domain.register(WelcomeEmail)
        test_domain.register(WelcomeNewPerson)

        spy = mocker.spy(SendgridEmailProvider, "send_email")

        Person.add_newcomer(
            {
                "email": "subhash.bhushan@gmail.com",
                "first_name": "John",
                "last_name": "Doe",
                "age": 21,
            }
        )
        assert spy.spy_return is True



# Standard Library Imports
from datetime import datetime

# Protean
from protean.core.aggregate import BaseAggregate
from protean.core.entity import BaseEntity
from protean.core.field.association import HasMany, HasOne, Reference
from protean.core.field.basic import DateTime, Integer, String, Text


class Post(BaseAggregate):
    title = String(required=True, max_length=1000)
    slug = String(required=True, max_length=1024)
    content = Text(required=True)
    posted_at = DateTime(required=True, default=datetime.now())

    meta = HasOne("PostMeta")
    comments = HasMany("Comment")


class PostMeta(BaseEntity):
    likes = Integer(default=0)

    post = Reference(Post)


class Comment(BaseEntity):
    content = Text(required=True)
    commented_at = DateTime(required=True, default=datetime.now())

    post = Reference(Post)

    class Meta:
        aggregate_cls = Post



# Standard Library Imports
from uuid import UUID

# Protean
import pytest

# Local/Relative Imports
from .elements import Person


class TestIdentityType:
    def test_string_identity(self, test_domain_with_string_identity):
        person = Person(first_name="John", last_name="Doe")
        assert person.id is not None
        assert isinstance(person.id, str)

    def test_int_identity(self, test_domain_with_int_identity):
        person = Person(first_name="John", last_name="Doe")
        assert person.id is not None
        assert isinstance(person.id, int)

    def test_uuid_identity(self, test_domain_with_uuid_identity):
        person = Person(first_name="John", last_name="Doe")
        assert person.id is not None

        try:
            UUID(str(person.id))
        except ValueError:
            pytest.fail("ID is not valid UUID")



# Protean
import pytest

from protean.globals import current_domain, g


class TestDomainContext:
    @pytest.fixture
    def test_domain(self):
        # Protean
        from protean.domain import Domain

        domain = Domain("Test")

        yield domain

    @pytest.fixture
    def test_domain_context(self, test_domain):
        with test_domain.domain_context() as context:
            yield context

    def test_domain_context_provides_domain_app(self, test_domain):
        with test_domain.domain_context():
            assert current_domain._get_current_object() == test_domain

    def test_domain_tearing_down(self, test_domain):
        cleanup_stuff = []

        @test_domain.teardown_domain_context
        def cleanup(exception):
            cleanup_stuff.append(exception)

        with test_domain.domain_context():
            pass

        assert cleanup_stuff == [None]

    def test_domain_tearing_down_with_previous_exception(self, test_domain):
        cleanup_stuff = []

        @test_domain.teardown_domain_context
        def cleanup(exception):
            cleanup_stuff.append(exception)

        try:
            raise Exception("dummy")
        except Exception:
            pass

        with test_domain.domain_context():
            pass

        assert cleanup_stuff == [None]

    def test_domain_tearing_down_with_handled_exception_by_except_block(
        self, test_domain
    ):
        cleanup_stuff = []

        @test_domain.teardown_domain_context
        def cleanup(exception):
            cleanup_stuff.append(exception)

        with test_domain.domain_context():
            try:
                raise Exception("dummy")
            except Exception:
                pass

        assert cleanup_stuff == [None]

    @pytest.mark.skip  # FIXME Could provide a domain level handler to catch and act on exceptions
    def test_domain_tearing_down_with_handled_exception_by_domain_handler(
        self, test_domain
    ):
        # app.config["PROPAGATE_EXCEPTIONS"] = True
        # cleanup_stuff = []

        # @test_domain.teardown_domain_context
        # def cleanup(exception):
        #     cleanup_stuff.append(exception)

        # @test_domain.errorhandler(Exception)
        # def handler(f):
        #     return flask.jsonify(str(f))

        # with domain.domain_context():
        #     client.get("/")

        # assert cleanup_stuff == [None]
        pass

    def test_domain_tearing_down_with_unhandled_exception(self, test_domain):
        test_domain.config["PROPAGATE_EXCEPTIONS"] = True
        cleanup_stuff = []

        @test_domain.teardown_domain_context
        def cleanup(exception):
            cleanup_stuff.append(exception)

        with pytest.raises(Exception):
            # Dummy class that is not registered with domain
            class FooBar:
                pass

            # Protean
            from protean.utils import DomainObjects

            with test_domain.domain_context():
                try:
                    test_domain._get_element_by_class(
                        (DomainObjects.AGGREGATE,), FooBar
                    )
                except Exception:
                    raise Exception("ElementNotFound")

        assert len(cleanup_stuff) == 1
        assert isinstance(cleanup_stuff[0], Exception)
        assert str(cleanup_stuff[0]) == "ElementNotFound"

    def test_domain_context_globals_methods(self, test_domain, test_domain_context):
        # get
        assert g.get("foo") is None
        assert g.get("foo", "bar") == "bar"
        # __contains__
        assert "foo" not in g
        g.foo = "bar"
        assert "foo" in g
        # setdefault
        g.setdefault("bar", "the cake is a lie")
        g.setdefault("bar", "hello world")
        assert g.bar == "the cake is a lie"
        # pop
        assert g.pop("bar") == "the cake is a lie"
        with pytest.raises(KeyError):
            g.pop("bar")
        assert g.pop("bar", "more cake") == "more cake"
        # __iter__
        assert list(g) == ["foo"]
        # __repr__
        assert repr(g) == "<protean.g of 'Test'>"

    def test_custom_domain_ctx_globals_class(self, test_domain):
        class CustomRequestGlobals:
            def __init__(self):
                self.spam = "eggs"

        test_domain.domain_context_globals_class = CustomRequestGlobals
        with test_domain.domain_context():
            assert g.spam == "eggs"



# Standard Library Imports
import logging

# Protean
from protean.core.exceptions import InvalidOperationError, ValidationError
from protean.globals import _uow_context_stack, current_domain

logger = logging.getLogger("protean.core.unit_of_work")


class UnitOfWork:
    def __init__(self):
        # Initialize session factories from all providers
        #   Connections will be retrieved at this stage

        # Also initialize Identity Map?
        #   Repository will first check here before retrieving from Database
        self.domain = current_domain
        self._in_progress = False

        self._sessions = {}
        self._events = []

    @property
    def in_progress(self):
        return self._in_progress

    def __enter__(self):
        # Initiate a new session as part of self
        self.start()
        return self

    def __exit__(self, *args):
        # Commit and destroy session
        self.commit()

    def start(self):
        # Stand in method for `__enter__`
        #   To explicitly begin and end transactions
        self._in_progress = True
        _uow_context_stack.push(self)

    def commit(self):
        # Raise error if there the Unit Of Work is not active
        logger.debug(f"Committing {self}...")
        if not self._in_progress:
            raise InvalidOperationError("UnitOfWork is not in progress")

        # Exit from Unit of Work
        _uow_context_stack.pop()

        # Commit and destroy session
        try:
            for _, session in self._sessions.items():
                session.commit()

            for event in self._events:
                for _, broker in self.domain.brokers.items():
                    broker.send_message(event)

            logger.debug("Commit Successful")
        except Exception as exc:
            logger.error(
                f"Error during Commit: {str(exc)}. Rolling back Transaction..."
            )
            self.rollback()
            raise ValidationError(
                {"_entity": [f"Error during Data Commit: - {repr(exc)}"]}
            )

        self._reset()

    def _reset(self):
        for _, session in self._sessions.items():
            session.close()

        self._sessions = {}
        self._events = []
        self._in_progress = False

    def rollback(self):
        # Raise error if the Unit Of Work is not active
        if not self._in_progress:
            raise InvalidOperationError("UnitOfWork is not in progress")

        # Exit from Unit of Work
        _uow_context_stack.pop()

        try:
            for _, session in self._sessions.items():
                session.rollback()

            logger.debug("Transaction rolled back")
        except Exception as exc:
            logger.error(f"Error during Transaction rollback: {str(exc)}")

        self._reset()

    def _get_session(self, provider_name):
        provider = self.domain.get_provider(provider_name)
        return provider.get_session()

    def _initialize_session(self, provider_name):
        new_session = self._get_session(provider_name)
        self._sessions[provider_name] = new_session
        if not new_session.is_active:
            new_session.begin()
        return new_session

    def get_session(self, provider_name):
        if provider_name in self._sessions:
            return self._sessions[provider_name]
        else:
            return self._initialize_session(provider_name)

    def register_event(self, event):
        self._events.append(event)



# Standard Library Imports
from abc import abstractmethod

# Protean
from protean import utils
from protean.core import exceptions
from protean.globals import current_domain

# Local/Relative Imports
from .base import Field
from .mixins import FieldCacheMixin, FieldDescriptorMixin
from .utils import fetch_entity_cls_from_registry


class _ReferenceField(Field):
    """Shadow Attribute Field to back References"""

    def __init__(self, reference, **kwargs):
        """Accept reference field as a an attribute, otherwise is a straightforward field"""
        self.reference = reference
        super().__init__(**kwargs)

    def __set__(self, instance, value):
        """Override `__set__` to update relation field and keep it in sync with the shadow
           attribute's value
        """
        value = self._load(value)

        if value:
            instance.__dict__[self.field_name] = value
        else:
            # Important to handle None assignment, and interpret it to mean resetting values
            self._reset_values(instance)

    def __delete__(self, instance):
        """Nullify values and linkages"""
        self._reset_values(instance)

    def _cast_to_type(self, value):
        """Verify type of value assigned to the shadow field"""
        # FIXME Verify that the value being assigned is compatible with the remote field
        return value

    def _reset_values(self, instance):
        """Reset all associated values and clean up dictionary items"""
        self.value = None
        self.reference.value = None
        instance.__dict__.pop(self.field_name, None)
        instance.__dict__.pop(self.reference.field_name, None)
        self.reference.delete_cached_value(instance)


class Reference(FieldCacheMixin, Field):
    """
    Provide a many-to-one relation by adding an attribute to the local entity
    to hold the remote value.

    By default ForeignKey will target the `id` column of the remote model but this
    behavior can be changed by using the ``via`` argument.
    """

    def __init__(self, to_cls, via=None, **kwargs):
        # FIXME ensure `via` argument is of type `str`
        super().__init__(**kwargs)
        self._to_cls = to_cls
        self.via = via

        self.relation = _ReferenceField(self)

    @property
    def to_cls(self):
        return self._to_cls

    def get_attribute_name(self):
        """Return attribute name suffixed with via if defined, or `_id`"""
        return "{}_{}".format(self.field_name, self.linked_attribute)

    def get_shadow_field(self):
        """Return shadow field
        Primarily used during Entity initialization to register shadow field"""
        return (self.attribute_name, self.relation)

    def get_cache_name(self):
        return self.field_name

    @property
    def linked_attribute(self):
        """Choose the Linkage attribute between `via` and designated `id_field` of the target class

        This method is initially called from `__set_name__()` -> `get_attribute_name()`
        at which point, the `to_cls` has not been initialized properly. We simply default
        the linked attribute to 'id' in that case.

        Eventually, when setting value the first time, the `to_cls` entity is initialized
        and the attribute name is reset correctly.
        """
        if isinstance(self.to_cls, str):
            return "id"
        else:
            return self.via or self.to_cls.meta_.id_field.attribute_name

    def _resolve_to_cls(self, instance):
        assert isinstance(self.to_cls, str)

        self._to_cls = fetch_entity_cls_from_registry(self.to_cls)

        # Refresh attribute name, now that we know `to_cls` Entity and it has been
        #   initialized with `id_field`
        self.attribute_name = self.get_attribute_name()

        # Reset the Shadow attribute's name
        self.relation = _ReferenceField(self)
        setattr(instance.__class__, self.attribute_name, self.relation)
        self.relation.__set_name__(instance, self.attribute_name)

        # Remove the earlier attribute if it is still attached
        old_attribute_name = "{}_{}".format(self.field_name, "id")
        if hasattr(instance, old_attribute_name):
            delattr(instance, old_attribute_name)

        # Update domain records because we enriched the class structure
        current_domain._replace_element_by_class(instance.__class__)

    def __get__(self, instance, owner):
        """Retrieve associated objects"""

        # If `to_cls` was specified as a string, take this opportunity to fetch
        #   and update the correct entity class against it, if not already done
        if isinstance(self.to_cls, str):
            self._resolve_to_cls(instance)

        reference_obj = None
        if hasattr(instance, "state_"):
            try:
                reference_obj = self.get_cached_value(instance)
            except KeyError:
                # Fetch target object by own Identifier
                id_field = self.get_attribute_name()

                id_value = None
                if hasattr(instance, id_field):
                    id_value = getattr(instance, id_field)

                if id_value:
                    reference_obj = self._fetch_objects(self.linked_attribute, id_value)
                    if reference_obj:
                        self._set_own_value(instance, reference_obj)
                    else:
                        # No Objects were found in the remote entity with this Entity's ID
                        pass

        return reference_obj

    def _fetch_objects(self, key, value):
        """Fetch Multiple linked objects"""
        return current_domain.get_dao(self.to_cls).find_by(**{key: value})

    def __set__(self, instance, value):
        """Override `__set__` to coordinate between relation field and its shadow attribute"""
        if value:
            # If `to_cls` was specified as a string, take this opportunity to fetch
            #   and update the correct entity class against it, if not already done
            if isinstance(self.to_cls, str):
                self._resolve_to_cls(instance)

            value = self._load(value)

            if value:
                # Check if the reference object has been saved. Otherwise, throw ValueError
                # FIXME not a comprehensive check. Should refer to state
                if getattr(value, value.meta_.id_field.field_name) is None:
                    raise ValueError(
                        "Target Object must be saved before being referenced",
                        self.field_name,
                    )
                else:
                    self._set_own_value(instance, value)
                    self._set_relation_value(
                        instance, getattr(value, self.linked_attribute)
                    )
        else:
            self._reset_values(instance)

    def _set_own_value(self, instance, value):
        self.value = value
        if value is None:
            instance.__dict__.pop(self.field_name, None)
            self.delete_cached_value(instance)
        else:
            instance.__dict__[self.field_name] = value
            self.set_cached_value(instance, value)

        # Mark Entity as Dirty
        if hasattr(instance, "state_"):
            instance.state_.mark_changed()

    def _set_relation_value(self, instance, value):
        self.relation.value = value
        if value is None:
            instance.__dict__.pop(self.attribute_name, None)
        else:
            instance.__dict__[self.attribute_name] = value

    def __delete__(self, instance):
        self._reset_values(instance)

    def _reset_values(self, instance):
        """Reset all associated values and clean up dictionary items"""
        self._set_own_value(instance, None)
        self._set_relation_value(instance, None)

    def _cast_to_type(self, value):
        # FIXME Assign value only of the correct type
        # if not isinstance(value, self.to_cls):
        #     self.fail('invalid', value=value)
        return value


class Association(FieldDescriptorMixin, FieldCacheMixin):
    """Base class for all association classes"""

    def __init__(self, to_cls, via=None, **kwargs):
        super().__init__(**kwargs)

        self.to_cls = to_cls
        self.via = via

        # FIXME Refactor for general use across all types of associations. Currently used only with `HasOne`
        self.change = None  # Used to store type of change in the association
        self.change_old_value = None  # Used to preserve the old value that was removed

        # Value holder
        self._value = None

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value if value else None

    def _cast_to_type(self, value):
        """Verify type of value assigned to the association field"""
        # FIXME Verify that the value being assigned is compatible with the associated Entity
        return value

    def _linked_attribute(self, owner):
        """Choose the Linkage attribute between `via` and own entity's `id_field`

           FIXME Explore converting this method into an attribute, and treating it
           uniformly at `association` level.
        """
        return self.via or (
            utils.inflection.underscore(owner.__name__)
            + "_"
            + owner.meta_.id_field.attribute_name
        )

    def __get__(self, instance, owner):
        """Retrieve associated objects"""

        # If `to_cls` was specified as a string, take this opportunity to fetch
        #   and update the correct entity class against it, if not already done
        if isinstance(self.to_cls, str):
            self.to_cls = fetch_entity_cls_from_registry(self.to_cls)

            # FIXME Test that `to_cls` contains a corresponding `Reference` field

        try:
            reference_obj = self.get_cached_value(instance)
        except KeyError:
            # Fetch target object by own Identifier
            id_value = getattr(instance, instance.meta_.id_field.field_name)
            reference_obj = self._fetch_objects(self._linked_attribute(owner), id_value)

            self._set_own_value(instance, reference_obj)

        return reference_obj

    def _set_own_value(self, instance, value):
        self.value = value
        instance.__dict__[self.field_name] = value
        self.set_cached_value(instance, value)

        # Mark Entity as Dirty
        if hasattr(instance, "state_"):
            instance.state_.mark_changed()

    @abstractmethod
    def _fetch_objects(self, key, value):
        """Placeholder method for customized Association query methods"""
        raise NotImplementedError

    def __set__(self, instance, value):
        """Cannot set values through an association"""
        raise exceptions.NotSupportedError(
            "Object does not support the operation being performed", self.field_name,
        )

    def __delete__(self, instance):
        """Cannot pop values for an association"""
        raise exceptions.NotSupportedError(
            "Object does not support the operation being performed", self.field_name,
        )

    def get_cache_name(self):
        return self.field_name

    @property
    def has_changed(self):
        return self.change is not None


class HasOne(Association):
    """
    Provide a HasOne relation to a remote entity.

    By default, the query will lookup an attribute of the form `<current_entity>_id`
    to fetch and populate. This behavior can be changed by using the `via` argument.
    """

    def __set__(self, instance, value):
        """Setup relationship to be persisted/updated"""
        # If `to_cls` was specified as a string, take this opportunity to fetch
        #   and update the correct entity class against it, if not already done
        if isinstance(self.to_cls, str):
            self.to_cls = fetch_entity_cls_from_registry(self.to_cls)

            # FIXME Test that `to_cls` contains a corresponding `Reference` field

        if value is not None:
            # This updates the parent's unique identifier in the child
            #   so that the foreign key relationship is preserved
            id_value = getattr(instance, instance.meta_.id_field.field_name)
            linked_attribute = self._linked_attribute(instance.__class__)
            setattr(
                value, linked_attribute, id_value
            )  # This overwrites any existing linkage, which is correct

        if self.value is None:
            self.change = "ADDED"
        elif value is None:
            self.change = "DELETED"
            self.change_old_value = self.value
        elif self.value.id != value.id:
            self.change = "UPDATED"
            self.change_old_value = self.value
        elif self.value.id == value.id and value.state_.is_changed:
            self.change = "UPDATED"
            self.change_old_value = self.value
        else:
            self.change = None  # The same object has been assigned, No-Op

        self._set_own_value(instance, value)

    def _fetch_objects(self, key, value):
        """Fetch single linked object"""
        try:
            return current_domain.get_dao(self.to_cls).find_by(**{key: value})
        except exceptions.ObjectNotFoundError:
            return None


class HasMany(Association):
    """
    Provide a HasMany relation to a remote entity.

    By default, the query will lookup an attribute of the form `<current_entity>_id`
    to fetch and populate. This behavior can be changed by using the `via` argument.
    """

    def _fetch_objects(self, key, value):
        """ Fetch linked entities.

        This method returns a well-formed query, containing the foreign-key constraint.
        The query will NOT be fired at this stage, and records will only be fetched
        when the consumer accesses the `HasMany` field values.

        The number of underlying entities fetched at one time is limited by the
        config attribute `AGGREGATE_CHILDREN_LIMIT`. By default, the value is `100`.
        Be aware that increasing this limit will mean more entity objects will be
        loaded into memory along with the aggregate.
        """
        children_dao = current_domain.get_dao(self.to_cls)
        return children_dao.query.filter(**{key: value}).limit(
            current_domain.config["AGGREGATE_CHILDREN_LIMIT"]
        )



# Protean
import pytest

from protean import Domain
from protean.core.exceptions import IncorrectUsageError
from protean.utils import fully_qualified_name

# Local/Relative Imports
from .elements import UserAggregate, UserEntity, UserFoo, UserVO


class TestDomainInitialization:
    def test_that_a_domain_can_be_initialized_successfully(self):
        domain = Domain(__name__)
        assert domain is not None
        assert domain.registry is not None
        assert domain.registry.aggregates == {}


class TestDomainRegistration:
    def test_that_only_recognized_element_types_can_be_registered(self, test_domain):

        with pytest.raises(NotImplementedError):
            test_domain.registry.register_element(UserFoo)

    def test_register_aggregate_with_domain(self, test_domain):
        test_domain.registry.register_element(UserAggregate)

        assert test_domain.registry.aggregates != {}
        assert fully_qualified_name(UserAggregate) in test_domain.registry.aggregates

    def test_register_entity_with_domain(self, test_domain):
        test_domain.registry.register_element(UserEntity)

        assert fully_qualified_name(UserEntity) in test_domain.registry.entities

    def test_register_value_object_with_domain(self, test_domain):
        test_domain.registry.register_element(UserVO)

        assert fully_qualified_name(UserVO) in test_domain.registry.value_objects

    def test_that_an_improperly_subclassed_element_cannot_be_registered(
        self, test_domain
    ):
        # Protean
        from protean.core.field.basic import String

        class Foo:
            pass

        class Bar(Foo):
            foo = String(max_length=50)

        with pytest.raises(NotImplementedError):
            test_domain.register(Bar)


class TestDomainAnnotations:
    # Individual test cases for registering domain elements with
    #   domain decorators are present in their respective test folders.

    def test_that_only_recognized_element_types_can_be_registered(self, test_domain):
        # Standard Library Imports
        from enum import Enum

        # Protean
        from protean.core.field.basic import String

        class DummyElement(Enum):
            FOO = "FOO"

        class FooBar:
            foo = String(max_length=50)

        with pytest.raises(IncorrectUsageError):
            test_domain._register_element(DummyElement.FOO, FooBar, aggregate_cls="foo")



"""Context Management Framework"""
# Standard Library Imports
import logging
import sys

# Protean
from protean.globals import _domain_context_stack

# a singleton sentinel value for parameter defaults
_sentinel = object()

logger = logging.getLogger("protean.application")


class _DomainContextGlobals(object):
    """A plain object. Used as a namespace for storing data during an
    domain context.

    Creating a domain context automatically creates this object, which is
    made available as the :data:`g` proxy.
    """

    def get(self, name, default=None):
        """Get an attribute by name, or a default value. Like
        :meth:`dict.get`.

        :param name: Name of attribute to get.
        :param default: Value to return if the attribute is not present.
        """
        return self.__dict__.get(name, default)

    def pop(self, name, default=_sentinel):
        """Get and remove an attribute by name. Like :meth:`dict.pop`.

        :param name: Name of attribute to pop.
        :param default: Value to return if the attribute is not present,
            instead of raise a ``KeyError``.
        """
        if default is _sentinel:
            return self.__dict__.pop(name)
        else:
            return self.__dict__.pop(name, default)

    def setdefault(self, name, default=None):
        """Get the value of an attribute if it is present, otherwise
        set and return a default value. Like :meth:`dict.setdefault`.

        :param name: Name of attribute to get.
        :param: default: Value to set and return if the attribute is not
            present.
        """
        return self.__dict__.setdefault(name, default)

    def __contains__(self, item):
        return item in self.__dict__

    def __iter__(self):
        return iter(self.__dict__)

    def __repr__(self):
        top = _domain_context_stack.top
        if top is not None:
            return "<protean.g of %r>" % top.domain.domain_name
        return object.__repr__(self)


def has_domain_context():
    """If you have code that wants to test if a domain context is there or
    not this function can be used.  You can also just do a boolean check on the
    :data:`current_domain` object instead.
    """
    return _domain_context_stack.top is not None


class DomainContext(object):
    """The domain context binds an domain object implicitly
    to the current thread or greenlet.
    """

    def __init__(self, domain):
        self.domain = domain
        self.g = domain.domain_context_globals_class()

        # Use a basic "refcount" to track number of domain contexts
        self._ref_count = 0

    def push(self):
        """Binds the domain context to the current context."""
        self._ref_count += 1
        if hasattr(sys, "exc_clear"):
            sys.exc_clear()
        _domain_context_stack.push(self)

    def pop(self, exc=_sentinel):
        """Pops the domain context."""
        try:
            self._ref_count -= 1
            if self._ref_count <= 0:
                if exc is _sentinel:
                    exc = sys.exc_info()[1]
                self.domain.do_teardown_domain_context(exc)
        finally:
            rv = _domain_context_stack.pop()
        assert rv is self, "Popped wrong domain context.  (%r instead of %r)" % (
            rv,
            self,
        )

    def __enter__(self):
        self.push()
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.pop(exc_value)

        if exc_type is not None:
            raise (exc_type, exc_value, tb)



# Standard Library Imports
import logging

from abc import abstractmethod

# Protean
from protean.core.exceptions import IncorrectUsageError
from protean.utils import DomainObjects, derive_element_class

logger = logging.getLogger("protean.domain.subscriber")


class _SubscriberMetaclass(type):
    """
    This base metaclass processes the class declaration and constructs a meta object that can
    be used to introspect the Subscriber class later. Specifically, it sets up a `meta_` attribute on
    the Subscriber to an instance of Meta, either the default of one that is defined in the
    Subscriber class.

    `meta_` is setup with these attributes:
        * `domain_event`: The domain_event that this subscriber is associated with
    """

    def __new__(mcs, name, bases, attrs, **kwargs):
        """Initialize Subscriber MetaClass and load attributes"""

        # Ensure initialization is only performed for subclasses of Subscriber
        # (excluding Subscriber class itself).
        parents = [b for b in bases if isinstance(b, _SubscriberMetaclass)]
        if not parents:
            return super().__new__(mcs, name, bases, attrs)

        # Remove `abstract` in base classes if defined
        for base in bases:
            if hasattr(base, "Meta") and hasattr(base.Meta, "abstract"):
                delattr(base.Meta, "abstract")

        new_class = super().__new__(mcs, name, bases, attrs, **kwargs)

        # Gather `Meta` class/object if defined
        attr_meta = attrs.pop("Meta", None)
        meta = attr_meta or getattr(new_class, "Meta", None)
        setattr(new_class, "meta_", SubscriberMeta(name, meta))

        return new_class


class SubscriberMeta:
    """ Metadata info for the Subscriber.

    Options:
    - ``domain_event``: The domain_event that this subscriber is associated with
    """

    def __init__(self, entity_name, meta):
        self.domain_event = getattr(meta, "domain_event", None)
        self.broker = getattr(meta, "broker", None)
        self.aggregate_cls = getattr(meta, "aggregate_cls", None)


class BaseSubscriber(metaclass=_SubscriberMetaclass):
    """Base Subscriber class that should implemented by all Domain Subscribers.

    This is also a marker class that is referenced when subscribers are registered
    with the domain
    """

    element_type = DomainObjects.SUBSCRIBER

    def __new__(cls, *args, **kwargs):
        if cls is BaseSubscriber:
            raise TypeError("BaseSubscriber cannot be instantiated")
        return super().__new__(cls)

    @classmethod
    @abstractmethod
    def notify(cls, domain_event):
        """Placeholder method for receiving notifications on domain event"""
        pass


def subscriber_factory(element_cls, **kwargs):
    element_cls = derive_element_class(element_cls, BaseSubscriber)

    element_cls.meta_.domain_event = (
        kwargs.pop("domain_event", None)
        or (hasattr(element_cls, "meta_") and element_cls.meta_.domain_event)
        or None
    )

    element_cls.meta_.broker = (
        kwargs.pop("broker", None)
        or (hasattr(element_cls, "meta_") and element_cls.meta_.broker)
        or "default"
    )

    if not element_cls.meta_.domain_event:
        raise IncorrectUsageError(
            f"Subscriber `{element_cls.__name__}` needs to be associated with a Domain Event"
        )

    if not element_cls.meta_.broker:
        raise IncorrectUsageError(
            f"Subscriber `{element_cls.__name__}` needs to be associated with a Broker"
        )

    return element_cls



""" Utility functions for the Field Module """
# Protean
from protean.globals import current_domain
from protean.utils import DomainObjects


def fetch_entity_cls_from_registry(entity):
    """Util Method to fetch an Entity class from an entity's name"""
    # Defensive check to ensure we only process if `to_cls` is a string
    if isinstance(entity, str):
        try:
            return current_domain._get_element_by_name(
                (DomainObjects.AGGREGATE, DomainObjects.ENTITY), entity
            ).cls
        except AssertionError:
            # Entity has not been registered (yet)
            # FIXME print a helpful debug message
            raise
    else:
        return entity


def fetch_value_object_cls_from_domain(value_object):
    """Util Method to fetch an Value Object class from a name string"""
    # Defensive check to ensure we only process if `value_object_cls` is a string
    if isinstance(value_object, str):
        try:
            return current_domain._get_element_by_name(
                DomainObjects.VALUE_OBJECT, value_object
            ).cls
        except AssertionError:
            # Value Object has not been registered (yet)
            # FIXME print a helpful debug message
            raise
    else:
        return value_object


def fetch_aggregate_cls_from_domain(aggregate):
    """Util Method to fetch an Aggregate class from a name string"""
    # Defensive check to ensure we only process if `aggregate_cls` is a string
    if isinstance(aggregate, str):
        try:
            return current_domain._get_element_by_name(
                DomainObjects.AGGREGATE, aggregate
            ).cls
        except AssertionError:
            # Aggregate has not been registered (yet)
            # FIXME print a helpful debug message
            raise
    else:
        return aggregate


def fetch_entity_cls_from_domain(entity):
    """Util Method to fetch an Entity class from a name string"""
    # Defensive check to ensure we only process if `entity_cls` is a string
    if isinstance(entity, str):
        try:
            return current_domain._get_element_by_name(DomainObjects.ENTITY, entity).cls
        except AssertionError:
            # Entity has not been registered (yet)
            # FIXME print a helpful debug message
            raise
    else:
        return entity



# Protean
import pytest

from protean.core.exceptions import ObjectNotFoundError

# Local/Relative Imports
from .elements import Person, PersonRepository


class TestAggregatePersistenceWithMemoryProvider:
    def test_retrieval_from_provider_connection(self, test_domain):
        conn = test_domain.get_connection()
        assert conn is not None

        conn._db["data"]["foo"] = "bar"

        assert "foo" in conn._db["data"]
        assert conn._db["data"]["foo"] == "bar"


class TestAggregatePersistenceWithRepository:
    @pytest.fixture(autouse=True)
    def register_repositories(self, test_domain):
        test_domain.register(Person)
        test_domain.register(PersonRepository)
        yield

    @pytest.fixture
    def person_dao(self, test_domain):
        return test_domain.get_dao(Person)

    def test_new_object_persistence_with_no_uow(self, test_domain):
        repo = test_domain.repository_for(Person)
        person = Person(first_name="John", last_name="Doe")
        repo.add(person)

        persisted_person = repo.get(person.id)
        assert persisted_person is not None

    def test_object_update_with_no_uow(self, test_domain):
        # Add a Person to the repository
        repo = test_domain.repository_for(Person)
        person = Person(first_name="John", last_name="Doe")
        repo.add(person)

        # Fetch person record back
        persisted_person = repo.get(person.id)
        assert persisted_person.last_name == "Doe"

        # Edit and update the person record
        persisted_person.last_name = "Dane"
        repo.add(persisted_person)

        # Re-fetch and check that the details have changed
        updated_person = repo.get(person.id)
        assert updated_person.last_name == "Dane"

    def test_object_delete_with_no_uow(self, test_domain):
        # Add a Person to the repository
        repo = test_domain.repository_for(Person)
        person = Person(first_name="John", last_name="Doe")
        repo.add(person)

        # Remove the person from repository
        repo.remove(person)

        # Re-fetch and check that the details have changed
        with pytest.raises(ObjectNotFoundError):
            repo.get(person.id)

    def test_that_an_aggregate_object_can_be_persisted_via_dao(self, person_dao):
        person = Person(first_name="John", last_name="Doe", age=35)
        persisted_person = person_dao.save(person)

        assert persisted_person is not None



# Protean
import pytest

from protean.utils.query import Q

# Local/Relative Imports
from .elements import Person


class TestConjunctions:
    """Class that holds tests cases for Conjunctions (AND, OR, NeG)"""

    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Person)

    @pytest.fixture
    def create_3_people(self, test_domain):
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

    @pytest.fixture
    def create_5_people(self, test_domain):
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )
        test_domain.get_dao(Person).create(
            id=5, first_name="Leslie", age=6, last_name="Underwood"
        )
        test_domain.get_dao(Person).create(
            id=6, first_name="Dave", age=6, last_name="Carrie"
        )

    def test_that_kwargs_to_filter_are_ANDed_by_default(
        self, test_domain, create_3_people
    ):
        q1 = test_domain.get_dao(Person).query.filter(last_name="John", age=3)
        assert q1.total == 1

    def test_that_kwargs_to_exclude_are_ORed_by_default(
        self, test_domain, create_3_people
    ):
        q1 = test_domain.get_dao(Person).query.exclude(last_name="John", age=3)
        assert q1.total == 1

        q2 = test_domain.get_dao(Person).query.exclude(last_name="Carrie", age=10)
        assert q2.total == 2

    def test_straightforward_AND_of_two_criteria(self, test_domain, create_3_people):
        # Filter by the Owner
        q1 = test_domain.get_dao(Person).query.filter(Q(last_name="John") & Q(age=3))
        assert q1.total == 1

    def test_straightforward_OR_of_two_criteria(self, test_domain, create_3_people):
        q1 = test_domain.get_dao(Person).query.filter(Q(last_name="John") | Q(age=3))
        assert q1.total == 2

    def test_combination_of_AND_followed_by_OR(self, test_domain, create_5_people):
        q1 = test_domain.get_dao(Person).query.filter(
            Q(last_name="John", first_name="Jean") | Q(age=6)
        )
        assert q1.total == 4

        q2 = test_domain.get_dao(Person).query.filter(Q(last_name="John") | Q(age=6))
        assert q2.total == 5

        q3 = test_domain.get_dao(Person).query.filter(
            (Q(last_name="John") & Q(age=7)) | (Q(last_name="Carrie") & Q(age=6))
        )
        assert q3.total == 3

    def test_combination_of_OR_followed_by_AND(self, test_domain, create_5_people):
        q1 = test_domain.get_dao(Person).query.filter(
            (Q(last_name="John") | Q(age=7)) & (Q(last_name="Carrie") | Q(age=6))
        )
        assert q1.total == 0

        q2 = test_domain.get_dao(Person).query.filter(
            (Q(last_name="John") | Q(age__gte=3))
            & (Q(first_name="Jean") | Q(first_name="Murdock"))
        )
        assert q2.total == 2

    def test_NEG_of_criteria(self, test_domain, create_5_people):
        q1 = test_domain.get_dao(Person).query.filter(~Q(last_name="John"))
        assert q1.total == 3

        q2 = test_domain.get_dao(Person).query.filter(~Q(last_name="John") | ~Q(age=7))
        assert q2.total == 4

    def test_empty_resultset_is_returned_correctly(self, test_domain, create_3_people):
        q1 = test_domain.get_dao(Person).query.filter(last_name="XYZ", age=100)
        assert q1.total == 0



""" Package for  Concrete Implementations of Protean repositories """
# Standard Library Imports
import importlib
import logging

# Protean
from protean.core.exceptions import ConfigurationError
from protean.utils import DomainObjects, fully_qualified_name

logger = logging.getLogger("protean.repository")


class Providers:
    def __init__(self, domain):
        self.domain = domain
        self._providers = None

    def get_model(self, aggregate_cls):
        """Retrieve Model class connected to Entity"""
        # Return model if already constructed
        if fully_qualified_name(aggregate_cls) in self.domain._constructed_models:
            return self.domain._constructed_models[fully_qualified_name(aggregate_cls)]

        # Fixate on the provider associated with the aggregate class
        aggregate_record = self.domain._get_element_by_class(
            (DomainObjects.AGGREGATE, DomainObjects.ENTITY), aggregate_cls
        )
        provider = self.get_provider(aggregate_record.cls.meta_.provider)

        # If a model was associated with the aggregate record, give it a higher priority
        #   and do not bake a new model class from aggregate/entity attributes
        custom_model_cls = None
        if fully_qualified_name(aggregate_cls) in self.domain._models:
            custom_model_cls = self.domain._models[fully_qualified_name(aggregate_cls)]

        # FIXME This is the provide support for activating database specific models
        #   This needs to be enhanced to allow Protean to hold multiple models per Aggregate/Entity
        #   per database.
        #
        #   If no database is specified, model can be used for all databases
        if custom_model_cls and (
            custom_model_cls.meta_.database is None
            or custom_model_cls.meta_.database == provider.conn_info["DATABASE"]
        ):
            # Get the decorated model class.
            #   This is a no-op if the provider decides that the model is fully-baked
            model_cls = provider.decorate_model_class(
                aggregate_record.cls, custom_model_cls
            )
        else:
            # No model was associated with the aggregate/entity explicitly.
            #   So ask the Provider to bake a new model, initialized properly for this aggregate
            #   and return it
            model_cls = provider.construct_model_class(aggregate_record.cls)

        self.domain._constructed_models[fully_qualified_name(aggregate_cls)] = model_cls
        return model_cls

    def _initialize_providers(self):
        """Read config file and initialize providers"""
        configured_providers = self.domain.config["DATABASES"]
        provider_objects = {}

        if configured_providers and isinstance(configured_providers, dict):
            if "default" not in configured_providers:
                raise ConfigurationError("You must define a 'default' provider")

            for provider_name, conn_info in configured_providers.items():
                provider_full_path = conn_info["PROVIDER"]
                provider_module, provider_class = provider_full_path.rsplit(
                    ".", maxsplit=1
                )

                provider_cls = getattr(
                    importlib.import_module(provider_module), provider_class
                )
                provider_objects[provider_name] = provider_cls(
                    provider_name, self, conn_info
                )

        return provider_objects

    def has_provider(self, provider_name):
        if self._providers is None:
            self._providers = self._initialize_providers()

        return provider_name in self._providers

    def get_provider(self, provider_name):
        """Retrieve the provider object with a given provider name"""
        if self._providers is None:
            self._providers = self._initialize_providers()

        try:
            return self._providers[provider_name]
        except KeyError:
            raise AssertionError(f"No Provider registered with name {provider_name}")

    def get_connection(self, provider_name="default"):
        """Fetch connection from Provider"""
        if self._providers is None:
            self._providers = self._initialize_providers()

        try:
            return self._providers[provider_name].get_connection()
        except KeyError:
            raise AssertionError(f"No Provider registered with name {provider_name}")

    def providers_list(self):
        """A generator that helps users iterator through providers"""
        if self._providers is None:
            self._providers = self._initialize_providers()

        for provider_name in self._providers:
            yield self._providers[provider_name]

    def repository_for(self, aggregate_cls):
        """Retrieve a Repository registered for the Aggregate"""
        # Protean
        from protean.core.aggregate import BaseAggregate

        if not issubclass(aggregate_cls, BaseAggregate):
            raise AssertionError(
                f"Element {aggregate_cls.__name__} must be subclass of `BaseAggregate`"
            )

        try:
            repository_record = next(
                repository
                for _, repository in self.domain.registry.repositories.items()
                if repository.cls.meta_.aggregate_cls.__name__ == aggregate_cls.__name__
            )
        except StopIteration:
            logger.debug(f"Constructing a Repository for {aggregate_cls}...")

            # Protean
            from protean.core.repository import BaseRepository

            new_class = type(
                aggregate_cls.__name__ + "Repository", (BaseRepository,), {}
            )

            self.domain._domain_element(
                DomainObjects.REPOSITORY, _cls=new_class, aggregate_cls=aggregate_cls,
            )

            # FIXME Avoid comparing classes / Fetch a Repository class directly by its aggregate class
            repository_record = next(
                repository
                for _, repository in self.domain.registry.repositories.items()
                if repository.cls.meta_.aggregate_cls.__name__ == aggregate_cls.__name__
            )

        return repository_record.cls()

    def get_dao(self, aggregate_cls):
        """Retrieve a DAO registered for the Aggregate with a live connection"""
        # Fixate on the provider associated with the aggregate class
        aggregate_record = self.domain._get_element_by_class(
            (DomainObjects.AGGREGATE, DomainObjects.ENTITY), aggregate_cls
        )
        provider = self.get_provider(aggregate_record.cls.meta_.provider)

        # Fixate on Model class at the domain level because an explicit model may have been registered
        model_cls = self.get_model(aggregate_record.cls)

        return provider.get_dao(aggregate_record.cls, model_cls)



# Protean
import pytest

from mock import patch
from protean.adapters import CeleryBroker
from protean.globals import current_domain
from tests.adapters.broker.celery_broker.elements import (
    NotifySSOSubscriber,
    Person,
    PersonAdded,
)


@pytest.mark.redis
class TestRedisConnection:
    def test_that_configured_broker_is_celery_with_redis(self):
        assert "default" in current_domain.brokers
        broker = current_domain.brokers["default"]

        assert isinstance(broker, CeleryBroker)
        assert broker.conn_info["URI"] == "redis://127.0.0.1:6379/2"
        assert broker.celery_app is not None


@pytest.mark.redis
class TestEventProcessing:
    @pytest.fixture(autouse=True)
    def register(self):
        current_domain.register(Person)
        current_domain.register(NotifySSOSubscriber)

    @patch.object(CeleryBroker, "send_message")
    def test_that_an_event_is_published_to_the_broker(self, mock):
        newcomer = Person.add_newcomer(
            {"first_name": "John", "last_name": "Doe", "age": 21}
        )
        mock.assert_called_once_with(
            PersonAdded(id=newcomer.id, first_name="John", last_name="Doe", age=21)
        )

    def test_that_events_are_available_on_queue_after_publish(self):
        Person.add_newcomer({"first_name": "John", "last_name": "Doe", "age": 21})



# Protean
from protean.core.domain_event import BaseDomainEvent
from protean.globals import current_domain
from protean.port.broker import BaseBroker
from protean.utils import fully_qualified_name


class InlineBroker(BaseBroker):
    def __init__(self, name, domain, conn_info):
        super().__init__(name, domain, conn_info)

        # In case of `InlineBroker`, the `IS_ASYNC` value will always be `False`.
        conn_info["IS_ASYNC"] = False

    def send_message(self, initiator_obj):
        if isinstance(initiator_obj, BaseDomainEvent):
            for subscriber in self._subscribers[
                fully_qualified_name(initiator_obj.__class__)
            ]:
                subscriber_object = subscriber(current_domain, initiator_obj.__class__)
                subscriber_object.notify(initiator_obj.to_dict())
        else:
            command_handler = self._command_handlers[
                fully_qualified_name(initiator_obj.__class__)
            ]
            command_handler.notify(initiator_obj.to_dict())



""" Define the send-grid email provider """
# Standard Library Imports
import logging

# Protean
from protean.core.email import BaseEmailProvider
from sendgrid import SendGridAPIClient
from sendgrid.helpers.mail import Mail, TemplateId

logger = logging.getLogger("protean.email.sendgrid")


class SendgridEmailProvider(BaseEmailProvider):
    """An email provider for sending emails using the Sendgrid Service"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sg_client = SendGridAPIClient(self.conn_info["API_KEY"])

    def send_email(self, message, dynamic_template=False):
        """ Send messages via the sendgrid api"""

        email = Mail(
            from_email=message.from_email or self.conn_info["DEFAULT_FROM_EMAIL"],
            to_emails=message.to,
        )
        email.dynamic_template_data = message.data
        email.template_id = TemplateId(message.template_id)

        try:
            response = self.sg_client.send(email)

            if response.status_code != 202:
                logger.error(
                    f"Error encountered while sending Email: {response.status_code}"
                )

            logger.debug(
                f"Email pushed to SendGrid successfully: {message.mime_message}"
            )
        except Exception as e:
            logger.error(f"Error encountered while sending Email: {e}")

        return True



# Protean
import pytest

from protean.core.domain_service import BaseDomainService
from protean.utils import fully_qualified_name

# Local/Relative Imports
from .elements import DummyDomainService


class TestDomainServiceInitialization:
    def test_that_base_domain_service_class_cannot_be_instantiated(self):
        with pytest.raises(TypeError):
            BaseDomainService()

    def test_that_domain_service_can_be_instantiated(self):
        service = DummyDomainService()
        assert service is not None


class TestDomainServiceRegistration:
    def test_that_domain_service_can_be_registered_with_domain(self, test_domain):
        test_domain.register(DummyDomainService)

        assert (
            fully_qualified_name(DummyDomainService)
            in test_domain.registry.domain_services
        )

    def test_that_domain_service_can_be_registered_via_annotations(self, test_domain):
        @test_domain.domain_service
        class AnnotatedDomainService:
            def special_method(self):
                pass

        assert (
            fully_qualified_name(AnnotatedDomainService)
            in test_domain.registry.domain_services
        )



# Standard Library Imports
from datetime import datetime
from enum import Enum

# Protean
import inflection
import pytest

from protean.core.aggregate import BaseAggregate
from protean.core.field.basic import DateTime, String
from protean.domain.registry import _DomainRegistry
from protean.utils import DomainObjects


class Role(BaseAggregate):
    name = String(max_length=15, required=True)
    created_on = DateTime(default=datetime.today())


def test_element_registration():
    register = _DomainRegistry()
    register.register_element(Role)

    assert (
        "tests.registry.tests.Role" in register._elements[DomainObjects.AGGREGATE.value]
    )


def test_delisting_element():
    register = _DomainRegistry()
    register.register_element(Role)

    assert (
        "tests.registry.tests.Role" in register._elements[DomainObjects.AGGREGATE.value]
    )

    register.delist_element(Role)

    assert (
        "tests.registry.tests.Role"
        not in register._elements[DomainObjects.AGGREGATE.value]
    )


def test_fetching_elements_from_registry():
    register = _DomainRegistry()
    register.register_element(Role)

    aggregates = register.get(DomainObjects.AGGREGATE)
    assert "tests.registry.tests.Role" in aggregates


def test_that_registry_exposes_properties():
    # Assert that lowercased, underscored, pluralized methods
    # are attached to registry for each element type
    assert all(
        inflection.pluralize(inflection.underscore(element_type.value.lower()))
        for element_type in DomainObjects
    )


def test_that_registering_an_unknown_element_type_triggers_an_error():
    class DummyEnum(Enum):
        UNKNOWN = "UNKNOWN"

    class FooBar1:
        element_type = "FOOBAR"

    class FooBar2:
        element_type = DummyEnum.UNKNOWN

    class FooBar3:
        pass

    register = _DomainRegistry()

    with pytest.raises(NotImplementedError):
        register.register_element(FooBar1)

    with pytest.raises(NotImplementedError):
        register.register_element(FooBar2)

    with pytest.raises(NotImplementedError):
        register.register_element(FooBar3)


def test_that_delisting_an_unknown_element_type_triggers_an_error():
    class DummyEnum(Enum):
        UNKNOWN = "UNKNOWN"

    class FooBar1:
        element_type = "FOOBAR"

    class FooBar2:
        element_type = DummyEnum.UNKNOWN

    class FooBar3:
        pass

    register = _DomainRegistry()

    with pytest.raises(NotImplementedError):
        register.delist_element(FooBar1)

    with pytest.raises(NotImplementedError):
        register.delist_element(FooBar2)

    with pytest.raises(NotImplementedError):
        register.delist_element(FooBar3)


def test_that_re_registering_an_element_has_no_effect():
    register = _DomainRegistry()
    register.register_element(Role)
    register.register_element(Role)

    assert len(register._elements[DomainObjects.AGGREGATE.value]) == 1
    assert (
        "tests.registry.tests.Role" in register._elements[DomainObjects.AGGREGATE.value]
    )



# Standard Library Imports
import importlib
import logging

from collections.abc import MutableMapping

# Protean
from protean.core.exceptions import ConfigurationError
from protean.globals import current_uow
from protean.utils import DomainObjects

logger = logging.getLogger("protean.broker")


class Brokers(MutableMapping):
    def __init__(self, domain):
        self.domain = domain
        self._brokers = None

    def __getitem__(self, key):
        if self._brokers is None:
            self._initialize()
        return self._brokers[key]

    def __iter__(self):
        if self._brokers is None:
            self._initialize()
        return iter(self._brokers)

    def __len__(self):
        if self._brokers is None:
            self._initialize()
        return len(self._brokers)

    def __setitem__(self, key, value):
        if self._brokers is None:
            self._initialize()
        self._brokers[key] = value

    def __delitem__(self, key):
        if self._brokers is None:
            self._initialize()
        del self._brokers[key]

    def _initialize(self):
        """Read config file and initialize brokers"""
        configured_brokers = self.domain.config["BROKERS"]
        broker_objects = {}

        logger.debug("Initializing brokers...")
        if configured_brokers and isinstance(configured_brokers, dict):
            if "default" not in configured_brokers:
                raise ConfigurationError("You must define a 'default' broker")

            for broker_name, conn_info in configured_brokers.items():
                broker_full_path = conn_info["PROVIDER"]
                broker_module, broker_class = broker_full_path.rsplit(".", maxsplit=1)

                broker_cls = getattr(
                    importlib.import_module(broker_module), broker_class
                )
                broker_objects[broker_name] = broker_cls(broker_name, self, conn_info)

        self._brokers = broker_objects

        # Initialize subscribers for Brokers
        for _, subscriber_record in self.domain.registry.subscribers.items():
            subscriber = subscriber_record.cls
            broker_name = subscriber.meta_.broker

            if broker_name not in self._brokers:
                raise ConfigurationError(
                    f"Broker {broker_name} has not been configured."
                )

            self._brokers[broker_name].register(
                subscriber.meta_.domain_event, subscriber
            )

        # Initialize command handlers for Brokers
        for _, command_handler_record in self.domain.registry.command_handlers.items():
            command_handler = command_handler_record.cls
            broker_name = command_handler.meta_.broker

            if broker_name not in self._brokers:
                raise ConfigurationError(
                    f"Broker {broker_name} has not been configured."
                )

            self._brokers[broker_name].register(
                command_handler.meta_.command_cls, command_handler
            )

    def publish(self, domain_event):
        """Publish a domain event to all registered brokers"""
        if self._brokers is None:
            self._initialize()

        # Log event into a table before pushing to brokers. This will give a chance to recover from errors.
        #   There is a pseudo-check to ensure `EventLog` is registered in the domain, to ensure that apps
        #   know about this functionality and opt for it explicitly.
        #   # FIXME Check if Event Log is enabled in config
        # Protean
        from protean.infra.event_log import EventLog

        if (
            "protean.infra.event_log.EventLog"
            in self.domain._domain_registry._elements[
                DomainObjects.AGGREGATE.value
            ]  # FIXME Do not refer to domain
        ):
            event_dao = self.domain.get_dao(EventLog)
            event_dao.create(
                kind=domain_event.__class__.__name__, payload=domain_event.to_dict()
            )

        if current_uow:
            logger.debug(
                f"Recording {domain_event.__class__.__name__} "
                f"with values {domain_event.to_dict()} in {current_uow}"
            )
            current_uow.register_event(domain_event)
        else:
            logger.debug(
                f"Publishing {domain_event.__class__.__name__} with values {domain_event.to_dict()}"
            )
            for broker_name in self._brokers:
                self._brokers[broker_name].send_message(domain_event)

    def publish_command(self, command):
        """Publish a command to registered command handler"""
        if self._brokers is None:
            self._initialize()

        if current_uow:
            logger.debug(
                f"Recording {command.__class__.__name__} "
                f"with values {command.to_dict()} in {current_uow}"
            )
            current_uow.register_command_handler(command)
        else:
            logger.debug(
                f"Publishing {command.__class__.__name__} with values {command.to_dict()}"
            )
            for broker_name in self._brokers:
                self._brokers[broker_name].send_message(command)



"""
Entrypoint module, in case you use `python -mprotean`.


Why does this file exist, and why __main__? For more info, read:

- https://www.python.org/dev/peps/pep-0338/
- https://docs.python.org/2/using/cmdline.html#cmdoption-m
- https://docs.python.org/3/using/cmdline.html#cmdoption-m
"""
# Protean
from protean.cli import main

if __name__ == "__main__":
    main()



# Protean
from protean.core.aggregate import BaseAggregate
from protean.core.domain_event import BaseDomainEvent
from protean.core.email import BaseEmail
from protean.core.exceptions import InsufficientDataError, InvalidDataError
from protean.core.field.basic import Integer, String
from protean.core.field.embedded import AggregateField
from protean.core.subscriber import BaseSubscriber
from protean.globals import current_domain


class Person(BaseAggregate):
    email = String(max_length=255, required=True)
    first_name = String(max_length=50, required=True)
    last_name = String(max_length=50, required=True)
    age = Integer(default=21)

    @classmethod
    def add_newcomer(cls, person_dict):
        """Factory method to add a new Person to the system"""
        newcomer = Person(
            email=person_dict["email"],
            first_name=person_dict["first_name"],
            last_name=person_dict["last_name"],
            age=person_dict["age"],
        )

        # Publish Event via the domain
        current_domain.publish(PersonAdded(person=newcomer))

        return newcomer


class PersonAdded(BaseDomainEvent):
    person = AggregateField(Person)


class WelcomeEmail(BaseEmail):
    """Emailer to welcome new additions"""

    SUBJECT = "Welcome to ABC!"
    TEMPLATE_ID = "d-2709347159b04353aae539cdf4b1217a"

    def __init__(self, to=None, data=None):
        if to is None or data is None:
            raise InsufficientDataError("`to` and `data` fields are mandatory")

        if not isinstance(data, dict):
            raise InvalidDataError({"data": ["should be a dict"]})

        super().__init__(
            subject=self.SUBJECT, template_id=self.TEMPLATE_ID, data=data, to=to
        )


class WelcomeNewPerson(BaseSubscriber):
    """Subscriber that notifies an external SSO system
    that a new person was added into the system
    """

    class Meta:
        domain_event = PersonAdded

    def notify(self, domain_event):
        email = WelcomeEmail(
            to=domain_event.person.email, data=domain_event.person.to_dict()
        )
        current_domain.send_email(email)
